% File encoding.pl 
% Encoding of fields in the torpedo puzzle
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(encoding, [init_encoding/2,
                     coded_field_neighbour/3,
                     group_count/4,
                     constrain_field/3,
                     empty_field_code/1,
                     coded_field_to_field/3,
                     label_board/2      
                    ]).

:- use_module(library(terms)).
:- use_module(library(clpfd)).
:- use_module(library(lists)).
:- use_module(environ).
:- use_module(utils).

% -------------------- The field encoding 

% In this encoding each field is represented by a single integer. A field
% can either be empty or a piece of a ship. In the latter case it is
% characterised by a 4-tuple <SDir,SLen,SPos,Clr>. For internal
% calculations we therefore introduce the data type `intern_field': 

%         Type               Definition                        Variable names
% ----------------------------------------------------------------------------

% :- type intern_field --->  /*empty:*/ e 
%                          ; /*ship: */  
%                            s(sdir,slength,spos,colour).       
%                            % ^ ship direction,                 
%                            %      ^ ship length,                      
%                            %              ^ position within the ship  
%                                                              % IField, IF
% :- type sdir --->          horiz ; vert.                     % SDir
% :- type slength ==         int.                              % SLen, SLenMax
% :- type spos ==            int.                              % SPos

% To carry out the encoding we will need two "maximal" values, the maximal
% colour number, ClrMax and the maximal ship length, SLenMax.

% The encoding is done in two steps. First, the set of allowed 3-tuples
% <SDir,SLen,SPos> is mapped into the interval 0..CKindMax, where
% CKindMax = SLenMax*(SLenMax+1)-2. This step is called the encoding of the
% kind. In the second step the encoded kind is combined with the colour to
% give the encoded field value.

% As an example, here is a table showing the encoding of kind for SLenMax=4
% (legend: v43 -->SDir=vert, SLen=4, SPos=3):

% 0:      h11=v11
 
% 1..4:   v44  h22 h32 h42
% 5..8:   v43 v33  h33 h43
% 9..12:  v42 v32 v22  h44

% 13..18: h21 v21 h31 v31 h41 v41

% The principle of this, somewhat arbitrary encoding is to to assign the
% highest set of codes to ship head pieces (i.e. those with position
% 1). Within the set of head pieces we assign higher codes to longer
% ships. This is so that max-style labeling will first place the longer
% ship heads on the board. The encoding of non-head pieces is not relevant,
% because, when a ship head is placed on the board, all other pieces of the
% ship are instantiated immediately.

% encode_kind(SDir, SLen, SPos, SLenMax, CKind):
% CKind is the kind-encoding of the ship piece <SDir,SLen,SPos>, given the
% maximal ship length SLenMax.
% :- type coded_kind == int.
% :- pred encode_kind(sdir::in, slength::in, spos::in, slength::in,
%                     coded_kind::out).
encode_kind(_, 1, 1, _, CKind) :- !, CKind = 0.
encode_kind(horiz, SLen, SPos, SLenMax, CKind) :-
        (   SPos >= 2 -> CKind is SLenMax*(SPos-2)+SLen
        ;   CKind is SLenMax*(SLenMax-1)+2*(SLen-1)-1
        ).
encode_kind(vert, SLen, SPos, SLenMax, CKind) :-
        (   SPos >= 2 ->
            CKind is SLenMax*(SLenMax-1)-SLenMax*(SPos-2)-SLen+1
        ;   CKind is SLenMax*(SLenMax-1)+2*(SLen-1)
        ).

% decode_kind(CKind, SLenMax, SDir, SLen, SPos):
% CKind is the kind-encoding of the ship piece <SDir,SLen,SPos>, given the
% maximal ship length SLenMax.
% :- pred decode_kind(coded_kind::in, slength::in,
%                     sdir::out, slength::out, spos::out).
decode_kind(0, _, SDir, SLen, SPos) :- !,
        SDir = horiz, SLen = 1, SPos = 1.
decode_kind(CKind, SLenMax, SDir, SLen, SPos) :-
        C1 is CKind-SLenMax*(SLenMax-1), C1 > 0, !,
        SLen is (C1-1)//2+2, SPos = 1,
        (   CKind mod 2 =:= 1 -> SDir = horiz
        ;   SDir = vert
        ).
decode_kind(CKind, SLenMax, SDir, SLen, SPos) :-
        SPos0 is (CKind-1)//SLenMax,
        SLen0 is (CKind-1) mod SLenMax,
        (   SPos0 >= SLen0 ->
            SDir = vert, SLen is SLenMax-SLen0, SPos is SLenMax-SPos0
        ;   SDir = horiz, SLen is SLen0+1, SPos is SPos0+2
        ).

% The second step of the field encoding is much simpler. The coded kind and
% the colour are combined so that the head pieces of longest ships of all
% colours are assigned the highest codes. The ship piece codes are mapped
% into the range 1..CodeMax, where CodeMax = (CKindMax+1)*ClrMax,  so that
% code 0 is reserved for the empty field.

% The CF coded field is the code of a piece with coded kind CKind and
% colour Clr, given the maximal colour number ClrMax.
% :- pred compose_coded_field(coded_kind::in, colour::in, colour::in,
%                             coded_field::out).
compose_coded_field(CKind, Clr, ClrMax, CF) :-
        CF is ClrMax*CKind+Clr.

% The CF coded field is the code of a piece with coded kind CKind and
% colour Clr, given the maximal colour number ClrMax.
% :- pred split_coded_field(coded_field::in, colour::in, 
%                           coded_kind::out, colour::out).
split_coded_field(CF, ClrMax, CKind, Clr) :-
        CKind is (CF-1)//ClrMax,
        Clr is CF-ClrMax*CKind.

% encode_field(IF, CF, SLenMax, ClrMax): CF is the code of internal field IF.
% :- pred encode_field(intern_field::in, coded_field::out,
%                      slength::in, colour::in).
encode_field(e, 0, _, _).
encode_field(s(SDir,SLen,SPos,Clr), CF, SLenMax, ClrMax) :-
        encode_kind(SDir, SLen, SPos, SLenMax, CKind),
        compose_coded_field(CKind, Clr, ClrMax, CF).

% decode_field(CF, IF, SLenMax, ClrMax): CF is the code of internal field IF. 
% :- pred decode_field(coded_field::in, intern_field::out,
%                      slength::in, colour::in).
decode_field(0, IField, _, _) :- !, IField = e.
decode_field(CF, s(SDir,SLen,SPos,Clr), SLenMax, ClrMax) :-
        split_coded_field(CF, ClrMax, CKind, Clr),
        decode_kind(CKind, SLenMax, SDir, SLen, SPos).

% CFieldMax is the maximal coded field value. 
% :- pred max_coded_field(coded_field::out, slength::in, colour::in).
max_coded_field(CFieldMax, SLenMax, ClrMax) :-
        encode_field(s(vert,SLenMax,1,ClrMax), CFieldMax, SLenMax, ClrMax).

% empty_field_code(CF): CF is the code of an empty field.
% :- pred empty_field_code(coded_field::out).
empty_field_code(0).


% -------------------- Encoding of groups

% We need to enumerate certain groups of coded fields, for use in
% group_count/4. For use in constrain_field/3, we also need to enumerate
% the coded fields named by a certain external field (e.g. those named by
% m-1: a middle piece of colour 1). We also need to get hold of a set of
% codes for some of the labeling methods. Naming for groups of codes is
% introduced by the following data type:

%         Type               Definition                        Variable names
% ----------------------------------------------------------------------------

% :- type group --->         empty
%                          ; head(slength)
%                          ; head(slength,colour)
%                          ; head(sdir,slength,colour)
%                          ; directed(sdir,colour)
%                          ; all(colour)
%                          ; field(field).                     % Group, G

% (For the definition of type field see torp_bed.pl.)

% CF is a coded field belonging to group G.
% :- pred group_coded_field(group::in, coded_field::out,
%                           slength::in, colour::in).
group_coded_field(G, CF, SLenMax, ClrMax) :-
        group_ifield(G, IField),
        any_ifield(IField, SLenMax, ClrMax),
        encode_field(IField, CF, SLenMax, ClrMax).

% any_ifield(IField, SLenMax, ClrMax): IField is an arbitrary valid
% internal field, given the maximal ship length SLenMax and the maximal
% colour number ClrMax.
% :- pred any_ifield(intern_field::out, slength::in, colour::in).
any_ifield(e, _, _).
any_ifield(s(SDir,SLen,SPos,Clr), SLenMax, ClrMax) :-
        any_dir(SDir),
        between(1, SLenMax, SLen),
        between(1, SLen, SPos),
        between(1, ClrMax, Clr).

% any_dir(SDir): SDir is an arbitrary ship direction.
% :- pred any_dir(sdir::out).
any_dir(horiz).
any_dir(vert).

% group_ifield(Group, IF): IF is an internal field belonging to group Group.
% :- pred group_ifield(group, intern_field).
% :- mode group_ifield(in, out).
% :- mode group_ifield(out, in).
group_ifield(empty,              e                       ).
group_ifield(head(SLen), 	 s(_SDir,SLen,1,_Clr)  	 ).
group_ifield(head(SLen,Clr), 	 s(_SDir,SLen,1,Clr)  	 ).
group_ifield(head(SDir,SLen,Clr),s(SDir,SLen,1,Clr)      ).
group_ifield(directed(SDir,Clr), s(SDir,SLen,_SPos,Clr)  ) :- dif(SLen,1).
group_ifield(all(Clr),           s(_SDir,_SLen,_SPos,Clr)).
group_ifield(field(=),           e                       ).
group_ifield(field(o-Clr),       s(horiz,1,1,Clr)        ).
group_ifield(field(u-Clr),       s(vert,SLen,1,Clr)      ) :- dif(SLen,1).
group_ifield(field(l-Clr),       s(horiz,SLen,1,Clr)     ) :- dif(SLen,1).
group_ifield(field(r-Clr),       s(horiz,SLen,SLen,Clr)  ) :- dif(SLen,1).
group_ifield(field(d-Clr),       s(vert,SLen,SLen,Clr)   ) :- dif(SLen,1).
group_ifield(field(m-Clr),       s(_SDir,SLen,SPos,Clr)  ) :- dif(SLen,1), 
        						      dif(SPos,1), 
        						      dif(SPos,SLen).

% -------------------- Group encoding storage and use

% To implement the group_count/4 and constrain_field/3 predicates, we need
% to convert a group name to the set of codes belonging to the group (or
% more exactly to the FDset of the codes).  Rather then to calculate this
% set at each usage, it is feasible to do the calculation once, at the
% start of solving a torpedo problem, and to store the results. For
% efficiency, this storage is implemented by compiling a file which is
% generated at initialisation time (cf. init_encoding/2 below).

% Given the max values SLenMax and ClrMax, generate all possible group names
% G and their associated FDsets, writing out clauses of form:
%           group_hash_fdset(GHash, G, FDset).
% where FDSet is the FDset of coded fields belonging to G and GHash is a
% hash code generated from G using term_hash/2 (library(terms). Hashing is
% used to speed up indexing.
% :- pred gen_groups(slength::in, colour::in).
gen_groups(SLenMax, ClrMax) :-
        setof(CF, group_coded_field(G, CF, SLenMax, ClrMax), CFs_of_G),
        list_to_fdset(CFs_of_G, FDSet_of_G),
        term_hash(G, GHash),
        format('group_hash_fdset(~q,~q,~q).~n',[GHash,G,FDSet_of_G]),
        fail.
gen_groups(_, _).

% group_count(Group, CFields, Count, Env): The number of coded fields which
% belong to group Group in the list CFields is constrained to be (the
% constraint variable) Count.
% :- pred group_count(group::in, coded_row::in, int::out, env::in).
group_count(Group, CFields, Count, Env) :-
        group_fdset(Group, FDSet, Env), !,
        fdset_membership(CFields, FDSet, Occs),
        sum(Occs, #=, Count).
group_count(_Group, _CFields, 0, _Env).

% FDSet is the set of coded fields belonging to Group.
% :- pred group_fdset(group::in, fdset::out, env::in).
group_fdset(Group, FDSet, _Env) :-
        term_hash(Group, GHash),
        group_hash_fdset(GHash, Group, FDSet).

% fdset_membership(Es, FDSet, Occs): Occs is a list of 0-1 values
% (constraint variables) reflecting the membership of the corresponding
% member of Es in the FDSet.
% :- pred fdset_membership(list(int)::in, fdset::in, list(int)::out).
fdset_membership([], _, []).
fdset_membership([E|Es], FDSet, [O|Occs]) :-
        O #<=> E in_set FDSet,
        fdset_membership(Es, FDSet, Occs).

% The coded field CF is constrained to belong to field Field.
% :- pred constrain_field(field::in, coded_field::in, env::in).
constrain_field(Field, CF, Env) :-       
        group_fdset(field(Field), FDSet, Env),
        CF in_set FDSet.

% The coded_field CF belongs to Field.
% type field2 ---> kind-colour ; = ; ' ' /*unknown*/ ; s /*some ship*/.
% :- pred coded_field_to_field(coded_field::in, field2::out).
coded_field_to_field(CF, Field, _Env) :-
        var(CF), !,
        (   fd_min(CF, 0) -> Field = ' '
        ;   Field = 's'
        ).
coded_field_to_field(CF, Field, Env) :-
        environment(maxes, Env, SLenMax, ClrMax),
        decode_field(CF, IField, SLenMax, ClrMax),
        group_ifield(field(Field), IField), !.

%-------------------- Horizontal and vertical neighbourship relations

% field_neighbour_range(CField, SDir, NbrRange, SLenMax, ClrMax):
% NbrRange is the range of coded field values that can neighbour CField in
% direction SDir. (CField is the left or upper neighbour.)
% There are two cases: an empty field has multiple neighbours, while a ship
% field has exactly one possible neighbour in the given direction.
% This predicate will be used to build a relation/3 goal describing the
% horizontal and vertical neighbourship relation.
% :- pred field_neighbour_range(coded_field::out, sdir::in, fdrange::out,
%                               slength::in, colour::in).
field_neighbour_range(0, Dir, EmptyNbrRange, SLenMax, ClrMax) :-
        findall(CField,
            neighbour_of_an_empty_field(Dir,CField,SLenMax,ClrMax), EmptyNbrs),
        list_to_fdset(EmptyNbrs, EmptyNbrFdSet),
        fdset_to_range(EmptyNbrFdSet, EmptyNbrRange).
field_neighbour_range(CField, Dir, {CFieldNbr}, SLenMax, ClrMax) :-
        any_ifield(Ship, SLenMax, ClrMax),
        ship_neighbour(Ship, Dir, IField),
        encode_field(Ship, CField, SLenMax, ClrMax),
        encode_field(IField, CFieldNbr, SLenMax, ClrMax).

% CField can be a neighbour of an empty field in direction Dir.
% :- pred neighbour_of_an_empty_field(sdir::in, coded_field::out,
%                                     slength::in, colour::in).
neighbour_of_an_empty_field(Dir, CField, SLenMax, ClrMax) :-
        any_ifield(IField, SLenMax, ClrMax),
        can_neighbour_empty_field(IField, Dir),
        encode_field(IField, CField, SLenMax, ClrMax).

% ship_neighbour(Ship, Dir, IField): The Ship internal field is a ship, and
% IField is its (right or down) neighbour in direction Dir.
% ship_neighbour(intern_field::in, sdir::in intern_field::out).
ship_neighbour(s(SDir,SLen,SPos,Clr), Dir, IField) :-
        (   SDir = Dir, SPos1 is SPos+1, SPos1 =< SLen ->
            IField = s(SDir,SLen,SPos1,Clr)
        ;   IField = e
        ).

% can_neighbour_empty_field(IField, Dir): IField can be a (right or down)
% neighbour of an empty field in direction Dir.
% :- pred can_neighbour_empty_field(intern_field::in, sdir::in).
can_neighbour_empty_field(e, _).
can_neighbour_empty_field(s(SDir,_SLen,SPos,_Clr), Dir) :-
        (   SDir = Dir -> SPos = 1
        ;   true
        ).

%-------------------- Diagonal neighbourship relations

% There are three implementation variants for the diagonal neighbourship,
% to be selected via an option.

% CF1 is a diagonal neighbour of CF2.
% :- pred diagonal_neighbour_reif(coded_field::in, coded_field::in).
diagonal_neighbour_reif(CF1, CF2) :-
        CF1 #= 0 #\/ CF2 #= 0.

% CF1 is a diagonal neighbour of CF2.
% :- pred diagonal_neighbour_arith(coded_field::in, coded_field::in).
diagonal_neighbour_arith(CF1, CF2) +:
        CF1 in 0 .. (1000-(min(CF2)/>1000)*1000),
        CF2 in 0 .. (1000-(min(CF1)/>1000)*1000).

% CF1 is a diagonal neighbour of CF2.
% :- pred diagonal_neighbour_cond(coded_field::in, coded_field::in).
diagonal_neighbour_cond(CF1, CF2) +:
        CF1 in (min(CF2)..0) ? (inf..sup) \/ {0},
        CF2 in (min(CF1)..0) ? (inf..sup) \/ {0}.


%-------------------- Initialisations

% Initalise the encoding module, initialise CFs, the list of all coded
% fields of the board.
% :- pred init_encoding(list(coded_field)::in, env::in).
init_encoding(CFs, Env) :-
        environment(maxes, Env, SLenMax, ClrMax),
        with_output_to_compiler(
           gen_clauses_to_be_compiled(SLenMax, ClrMax, Env)),
        max_coded_field(CFMax, SLenMax, ClrMax),
        domain(CFs, 0, CFMax).

% Output (for compilation) the clauses describing the neighbourship
% relations and group encoding.
% :- pred gen_clauses_to_be_compiled(slength::in, colour::in, env::in).
gen_clauses_to_be_compiled(SLenMax, ClrMax, Env) :-
        gen_neighbour_switch(Env),
	option(relation(Rel), Rel=indexical, Env),
        gen_neighbour(horiz, Rel, SLenMax, ClrMax),
        gen_neighbour(vert, Rel, SLenMax, ClrMax),
        gen_groups(SLenMax, ClrMax).

% Output the three clauses for coded_field_neighbour/3. Select the
% applicable diagonal constraint using the option diag(...).
% :- pred gen_neighbour_switch(env::in).
gen_neighbour_switch(Env) :-
        option(diag(Diag), Diag=ind_cond, Env),
        diag_constraint_option(Diag, PredName),
        format('coded_field_neighbour(diag,CF,CFN) :- \n\t~w(CF,CFN).\n',
               [PredName]),
        format('coded_field_neighbour(horiz,CF,CFN) :- \n', []),
        format(                    '\thoriz_neighbour(CF,CFN).\n',  []),
        format('coded_field_neighbour(vert,CF,CFN) :- \n', []),
        format(                    '\tvert_neighbour(CF,CFN).\n',  []).

% diag_constraint_option(Option, PredName): PredName corresponds to diag
% option value Option.
% :- type diag_option ---> reif ; ind_arith ; ind_cond.
% :- pred diag_constraint_option(diag_option::in, atom::out).
diag_constraint_option(reif, diagonal_neighbour_reif).
diag_constraint_option(ind_arith, diagonal_neighbour_arith).
diag_constraint_option(ind_cond, diagonal_neighbour_cond).

% Output the definition of a neighbourship relation in direction Dir.
% Use the format indicated by `relation' option Rel: Rel = indexical
% means generate an indexical with relation/3 in the body; Rel = clause
% means generate a plain clause.
% :- type rel_option ---> indexical ; clause.
% :- pred gen_neighbour(sdir::in, rel_option::in, slength::in, colour::in).
gen_neighbour(Dir, Rel, SLenMax, ClrMax) :-
        findall(CF-NbrRange,
                field_neighbour_range(CF, Dir, NbrRange, SLenMax, ClrMax),
                NbrRel),
	(   Rel = indexical -> Neck = '+:'
	;   Neck = ':-'
	),
        format('~a_neighbour(CF,CFN) ~a\n   relation(CF,~w,CFN).~n',
               [Dir,Neck,NbrRel]).

%-------------------- Labeling

% We have an option label(LabOpt) deciding on what labeling to use:

% :- type labeling_option --->    none ; plain ; max_dual ; ships.

% The `none' case means no labeling is done, this is useful for checking
% how much information is deduced by the constraints without labeling.

% The `plain' labeling option calls the built in labeling/2 with the
% [max,down] settings. 

% The `max_dual' labeling performs a dual labeling, derived from the
% squares packing example of Hentenryck (?). This goes through the coded
% field values from the highest to the lowest and keeps trying to assign
% the highest possible values to the variables.

% Finally the `ships' labeling goes through the all ship lengths, from
% longest to the shortest, and for each length goes through all the colours
% and tries to position the (vertical or horizontal) ships of the given
% length and colour. This seems to be the best labeling strategy for the
% more complex examples.

% Within labeling we have an optional filtering phase, a kind of
% constructive disjunction handling. The filter_vars predicate scans all
% variables and checks if instantiating them to 0 (empty field code) causes
% failure. If so, the value 0 is excluded from the variable domain,
% otherwise the variable is left unchanged. Filtering is done at the start
% of the labeling phase and in each outer loop of max_dual labeling.

% The filter option can have the following values

% :- type filter_options ==  list(filter_option).
% :- type filter_option ---> off ; on ; repetitive ;
%                            verbose

% The value `off' means no filtering, `on' (the default) means a single
% scan of the variable list is made, while the `repetitive' option
% indicates that a repeated scan is to be made if variables got
% filtered. The verbose option triggers the output of filtering statistics.

% The variables in coded board CBd are labeled.
% :- pred label_board(coded_board::in, env::in).
label_board(CBd, Env) :-
        append_lists(CBd, CFsAll),
        option(filter(FiltOpts), FiltOpts=[], Env),
        filter_vars(FiltOpts, CFsAll, CFsRest),
        option(label(LabOpt), LabOpt=ships, Env),
        label_board(LabOpt, CFsRest, FiltOpts, Env).

% label_board(LabOpt, CFs, FiltOpts, Env): The elements of CFs are labeled
% according to LabOpt, with filtering depending on FiltOpts.
% :- pred label_board(labeling_option::in, list(coded_field)::in,
% 		      filter_options::in, env::in).
label_board(none, _CFs, _FiltOpts, _Env) :- !.
label_board(plain, CFs, _FiltOpts, _Env) :- !,
        labeling([max,down], CFs).
label_board(max_dual, CFs, FiltOpts, Env) :- !,
        environment(maxes, Env, SLenMax, ClrMax),
        max_coded_field(CFMax, SLenMax, ClrMax),
        max_dual_labeling(CFs, FiltOpts, CFMax).
label_board(ships, CFs, FiltOpts, Env) :- !,
	environment(ship_length_max, Env, SLenMax),
	ship_labeling(SLenMax, CFs, FiltOpts, Env).
label_board(LabOpt, _, _, _) :-
	format(user_error, '{ERROR: unknown labeling option: ~w}~n', [LabOpt]),
        fail.	

%-------------------- Filtering variables

% filter_vars(FiltOpts, Vars0, Vars): Filter variables Vars0, according to
% FiltOpts, resulting in Vars.
% :- pred filter_vars(filter_options::in, list(var)::in, list(var)::out).
filter_vars(FiltOpts, Vars, Vars) :- 
	member(off, FiltOpts), !.
filter_vars(FiltOpts, Vars0, Vars) :-
        statistics(runtime, _),
        filter_count_vars(Vars0, Vars1, 0, S),
        (   member(verbose, FiltOpts) ->
            statistics(runtime, [_,T]),
            format('Filtering: ~d vars, time: ~3d~n', [S,T])
        ;   true
        ),
        (   member(repetitive, FiltOpts),  S > 0 -> 
            filter_vars(FiltOpts, Vars1, Vars)
        ;   Vars = Vars1
        ).

% filter_count_vars(Vars0, Vars, Cnt0, Cnt): Filter variables Vars0,
% resulting in Vars. The number of variables filtered is Cnt-Cnt0.
% :- pred filter_count_vars(list(var)::in, list(var)::out, int::in, int::out).
filter_count_vars([], [], Cnt, Cnt).
filter_count_vars([V|Vs], Fs, Cnt0, Cnt) :-
        integer(V), !,
        filter_count_vars(Vs, Fs, Cnt0, Cnt).
filter_count_vars([V|Vs], [V|Fs], Cnt0, Cnt) :-
        (   fd_min(V, Min), Min > 0 -> Cnt1 = Cnt0
        ;   \+ (V = 0) -> V #\= 0, Cnt1 is Cnt0+1
        ;   Cnt1 = Cnt0
        ),
        filter_count_vars(Vs, Fs, Cnt1, Cnt).

%-------------------- Dual maximum labeling

% max_dual_labeling(Vars, FiltOpts, Max): Label Vars by assigning highest
% values first, given the maximal possible value is Max. Do a filtering
% whenever Max decreases, according to FiltOpts.
% :- pred max_dual_labeling(list(var)::in, filter_options::in, int::in).
max_dual_labeling([], _, _) :- !.
max_dual_labeling(Vars0, FiltOpts, Max) :-
        filter_vars(FiltOpts, Vars0, Vars),
        max_dual_labeling(Vars, Vars1, Max, 0, NMax),
        max_dual_labeling(Vars1, FiltOpts, NMax).

% max_dual_labeling(Vars0, Vars, Max, NewMax0, NewMax): Label Vars0 by
% trying to assign Max to as many variables as possible. Vars is the list
% of remaining uninstantiated variables, NewMax is the maximum of the
% highest values in domains of Vars and of NewMax0.
% :- pred max_dual_labeling(list(var)::in, list(var)::out, int::in,
%                           int::in, int::out).
max_dual_labeling([], [], _, NMax, NMax).
max_dual_labeling([V|Vars1], Vars2, Max, NMax0, NMax) :-
        (   integer(V) -> max_dual_labeling(Vars1, Vars2, Max, NMax0, NMax)
        ;   V = Max,
            max_dual_labeling(Vars1, Vars2, Max, NMax0, NMax)
        ;   V #< Max,
            fd_max(V, NMax1), NMax2 is max(NMax0,NMax1), Vars2 = [V|Vars3],
            max_dual_labeling(Vars1, Vars3, Max, NMax2, NMax)
        ).

%-------------------- Special ship labeling

% ship_labeling(SLen, Vs, FiltOpts, Env): Label Vs, which is guaranteed to
% contain only ship pieces with length SLen and below.
% :- pred ship_labeling(slen::in, list(coded_field)::in,
%                       filter_options::in, env::in).
ship_labeling(1, Vs, _, _) :-
	!, labeling([ff], Vs).
ship_labeling(SLen, Vs0, FiltOpts, Env) :-
	environment(colour_max, Env, ClrMax),
	ship_labeling(ClrMax, SLen, Vs0, Vs1, FiltOpts, Env),
	SLen1 is SLen-1,
	ship_labeling(SLen1, Vs1, FiltOpts, Env).

% ship_labeling(Clr, SLen, Vs0, Vs, FiltOpts, Env): Label the variables in
% Vs0 with ship pieces the length of which is SLen and the colour of which
% is less or equal to Clr. Vs is the list of remaining, uninstantiated
% variables.
% :- pred ship_labeling(colour::in, slen::in, list(coded_field)::in,
%                       list(coded_field)::out, filter_options::in, env::in).
ship_labeling(0, _, Vs, Vs, _, _) :- !.
ship_labeling(Clr, SLen, Vs0, Vs, FiltOpts, Env) :-
	label_group(head(SLen,Clr), Vs0, Vs1, FiltOpts, Env),
	Clr1 is Clr-1,
	ship_labeling(Clr1, SLen, Vs1, Vs, FiltOpts, Env).

% Label the variables in Vs0 with ship pieces belonging to Group. Vs is the
% list of remaining, uninstantiated variables.
% :- pred label_group(group::in, list(coded_field)::in,
%                     list(coded_field)::out, filter_options::in, env::in).
label_group(Group, Vs0, Vs, FiltOpts, Env) :-
	filter_vars(FiltOpts, Vs0, Vs1),
	group_fdset(Group, Set, Env),
	fdset_complement(Set, ComplSet),
	label_fdset(Vs1, Set, ComplSet, Vs).

% label_fdset(Vs, Set, ComplSet, UVs): Label the variables in Vs with
% values belonging to fd_set Set. ComplSet is the complement of Set, UVs is
% the list of remaining, uninstantiated variables.
% :- pred label_fdset(list(var)::in, fdset::in, fdset::in, list(var)::out).
label_fdset([], _Set, _ComplSet, []).
label_fdset([V|Vs], Set, ComplSet, UVs) :-
	integer(V), !, label_fdset(Vs, Set, ComplSet, UVs).
label_fdset([V|Vs], Set, ComplSet, UVs) :-
	fd_set(V, VSet), fdset_intersection(VSet, Set, Intersec),
	(   empty_fdset(Intersec) -> UVs = [V|UVs1]
	;   fdset_member(V, Intersec), UVs = UVs1
	;   V in_set ComplSet, UVs = [V|UVs1]
	),
	label_fdset(Vs, Set, ComplSet, UVs1).
