% File environ.pl
% Environment and option handling for the torpedo puzzle
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(environment, [
			make_environment/6,
			environment/3,
			environment/4,
			option/3,
			option/2,
			constraint_option/2
		       ]).

:- use_module(library(lists)).
:- use_module(utils).

%-------------------- The environment data structure

% The environment data structure stores the board size, the maximal ship
% length (SLenMax) and colour values and the options. It also stores the
% list [1,...,SLenMax].

% environment --->      e(row_ind, clm_ind, slength, colour,
%                         list(slength), options). 		% Env

% Build the environment Env of the given max-values an options).
% :- pred make_environment(row_ind::in, clm_ind::in, slength::in,
%                          colour::in, options::in, env::out).
make_environment(RowNum, ClmNum, SLenMax, ClrMax, Opts, Env) :-
	sequence(1, SLenMax, SLenSeq),
	Env = e(RowNum,ClmNum,SLenMax,ClrMax,SLenSeq,Opts).

% environment(Sel, Env, Max1, Max2): The values Max1 and Max2 are the
% components of Env selected by the selector Sel.
% :- type selector == atom.
% :- pred environment(selector::in, env::in, any::out, any::out).
environment(sizes, e(RowNum,ClmNum,_,_,_,_), RowNum, ClmNum).
environment(maxes,e(_,_,SLenMax,ClrMax,_,_), SLenMax, ClrMax).

% environment(Sel, Env, Val): The value Val is the component of Env
% selected by the selector Sel. 
% :- pred environment(selector::in, env::in, any::out).
environment(ship_length_max, e(_,_,SLenMax,_,_,_), SLenMax).
environment(colour_max,e(_,_,_,ClrMax,_,_), ClrMax).
environment(ship_length_seq, e(_,_,_,_,SLenSeq,_), SLenSeq).
environment(options, e(_,_,_,_,_,Opts), Opts).

% option(Opt, Default, Env): Opt is either present in the options component
% of Env, or Default is executed.
% :- pred option(options::in, any::in, env::in).
option(Opt, _Default, Env) :-
	environment(options, Env, Opts),
	member(Opt, Opts), !.
option(_, Default, _) :-
	Default.

% option(Opt, Env): Opt is present in the options component of Env.
% :- pred option(options::in, env::in).
option(Opt, Env) :-
	environment(options, Env, Opts),
	member(Opt, Opts), !.

% All redundant constraints are enabled by default. To disable some of
% these, include these constraint names prefixed by -. Alternatively, if
% you want to specify which constraints to enable, include their names
% prefixed by a +. To disable all redundant constraints include e.g. the
% +none option.

% constraint_option(Constraint, Env): If there are any +C format options in
% Env, then there is a +Constraint % option in Env, otherwise there is no
% -Constraint option in Env.
% :- pred constraint_option(atom::in, env::in).
constraint_option(Constraint, Env) :-
	option(+_, Env), !, option(+Constraint, Env).
constraint_option(Constraint, Env) :-
	option(-Constraint, Env), !, fail.
constraint_option(_, _).




