% File torp.pl
% The topmost layer of the torpedo puzzle solver
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(torpedo, [torpedo/2, torpedo/3]).

% :- export torpedo/2, torpedo/3.
% :- pred torpedo(torp::in, board::out).
% :- pred torpedo(torp::in, board::out, options::in).

:- use_module(library(clpfd)).
:- use_module(library(lists)).

% An overview of program structure
% --------------------------------

% This module contains the top layer of the torpedo constraint problem
% solver. It assumes that the board to be constructed is stored as a
% matrix, the elements of which are of type `coded_field':

% coded_board ==        list(coded_row).	% CBd, CBdT /*transposed*/,
%						% CRows
% coded_row == 		list(coded_field).	% CRow, CFs
% coded_field        	/* see encode.pl */	% CField, CF

% The `coded_field' type is defined in the module `encoding', which can
% choose to implement it as a single constraint variable, or as a structure
% composed of several constraint variables. The `torpedo' module uses
% `coded_field' as an abstract data type, through the operations provided
% by the `encoding' module.  However, the encoding is assumed to be
% sufficiently fine, so that all four characteristics of a ship piece can
% be distinguished: colour (Clr), direction (SDir --- `horiz' or `vert'),
% length (SLen) and position within the ship (SPos).

% The basic constraints provided by the `encoding' module are:

%         coded_field_neighbour(Dir, CField, CFieldNeighbour):
%             CFieldNeighbour is constrained to be a neighbour of CField in the
%             direction Dir (which can be `horiz', `vert', or `diag').

%         group_count(Group, CFields, Count, Env):
%             The number of coded fields which belong to group Group in the
%             list CFields is constrained to be (the constraint variable)
%             Count. Here Group can be one of:
%                 empty               --- an empty field
%                 head(SDir,SLen,Clr) --- a head field (uppermost or
%                                         leftmost) of a ship with a given
%                                         direction, length and colour
%                 directed(SDir,Clr)  --- any field of a directed ship
%                                         (longer than 1) with a given
%                                         direction and colour.
%		  all(Clr)            --- any field of a ship with a given
%                                         colour.
%             For the role of the environment Env, see below.

%         constrain_field(Field, CField, Env):
%             The coded field CField is constrained to contain the given
%             field Field (which can be empty: =, or a ship piece: Kind-Clr).

% The `encoding' module also deals with labeling (label_board/2), as this
% operation very much depends on the encoding. Further auxiliary predicates
% provided by the `encoding' module are: init_encoding/2,
% empty_field_code/1 (returns the coded field representation of empty) and
% coded_field_to_field/3 (converts a coded field to the field format
% required on output).

% Throughout most of the program we carry around an `environment' argument,
% which contains the global parameters of the torpedo problem, such as the
% board size, number of colours, etc. The actual format of this environment
% structure is defined in the `environment' module. This module also
% provides some option handling primitives: options are used to enable or
% disable certain redundant constraints.

:- use_module(encoding).
:- use_module(environ).
:- use_module(utils).
:- use_module(torp_utils, [collapsed_colours_descr/2]).

% torpedo(TDescr, Bd): Bd is a solution of of the torpedo problem TDescr.
% :- pred torpedo(torp_descr::in, board::out).
torpedo(TDescr, Bd) :-
	torpedo(TDescr, Bd, []).

% torpedo(TDescr, Bd, Opts): Bd is a solution of of the torpedo
% problem TDescr, found using options Opts.
% :- pred torpedo(torp_descr::in, board::out, options::in).
torpedo(TDescr, Bd, Opts) :-
	TDescr = descr(KnownEs, ClrDescrs),
	init_board(ClrDescrs, Opts, CBd, CBdT, Env),
	constrain_empties(ClrDescrs, KnownEs, CBd, CBdT, Env),
	constrain_colours(ClrDescrs, 1, CBd, CBdT, Env),
	constrain_layout(CBd, Env),
	label_board(CBd, Env),
	coded_board_to_board(CBd, Bd, Env).

% ----------------------------------------
% Initialisations

% CBd is a coded board corresponding to ClrDescrs, and CBdT is its transposed
% form. The environment Env is initialised from ClrDescrs and Opts.
% :- pred init_board(colours_descr::in, options::in,
%                    coded_board::out, coded_board::out, env::out).
init_board(ClrDescrs, Opts, CBd, CBdT, Env) :-
	length(ClrDescrs, ClrMax),
	ClrDescrs = [c(_,RowOccs,ClmOccs,_)|_],
	length(RowOccs, RowNum),
	length(ClmOccs, ClmNum),
	max_ship_length(ClrDescrs, 0, SLenMax),
	make_environment(RowNum, ClmNum, SLenMax, ClrMax, Opts, Env),
	matrix_of_vars(RowNum, ClmNum, CBd),
	transpose(CBd, CBdT),
	append_lists(CBd, CFsAll),
	init_encoding(CFsAll, Env).

% max_ship_length(ClrDescrs, Max0, Max): the maximum of the maximal ship
% length in ClrDescrs and Max0 is Max.
% :- pred max_ship_length(colours_descr::in, int::in, int::out).
max_ship_length([], Max, Max).
max_ship_length([c(ShCnts,_,_,_)|ClrDescrs], Max0, Max) :-
	length(ShCnts, Max1),
	Max2 is max(Max0,Max1),
	max_ship_length(ClrDescrs, Max2, Max).

% ----------------------------------------
% Constraints related to empty fields

% The coded board CBd is constrained to contain empty fields at positions
% prescribed by KnownEs. As an optional redundant constraint, each row and
% column of the board (CBd and CBdT) is constrained to contain a prescribed
% number of empty pieces (any field not containing a coloured piece, as
% described by counts in ClrDescrs can be taken to be empty).
% :- pred constrain_empties(colours_descr::in, known_empties::in,
%                           coded_board::in, coded_board::in, env::in).
constrain_empties(ClrDescrs, KnownEs, CBd, CBdT, Env) :-
	constrain_known_empties(KnownEs, CBd, Env),
	(   constraint_option(count_empties, Env) ->
	    collapsed_colours_descr(ClrDescrs, c(_,RowOccsSum,ClmOccsSum,_)),
	    environment(sizes, Env, RowNum, ClmNum),
	    constrain_empty_counts(CBd, RowOccsSum, ClmNum, Env),
	    constrain_empty_counts(CBdT, ClmOccsSum, RowNum, Env)
	;   true
	).

% constrain_known_empties(KnownEs, CBd, Env): The coded board CBd is
% constrained to contain empty fields at positions prescribed by KnownEs.
% :- pred constrain_known_empties(known_empties::in, coded_board::in, env::in).
constrain_known_empties([], _, _).
constrain_known_empties([RowI-ClmI|KnownEs], CBd, Env) :-
	matrix_elem(RowI, ClmI, CBd, CF),
	constrain_field(=, CF, Env),
	constrain_known_empties(KnownEs, CBd, Env).

% constrain_empty_counts(CRows, RowOccs, ClmNum, Env): Each coded_row in
% CRows is constrained to contain ClmNum-RowOcc empty fields, where RowOcc
% is the corresponding element of RowOccs.
% :- pred constrain_empty_counts(coded_board::in, row_occs::in,
%                                column_index::in, env::in).
constrain_empty_counts([], [], _ClmNum, _Env).
constrain_empty_counts([CRow|CRows], [RowOcc|RowOccs], ClmNum, Env) :-
	Empties is ClmNum-RowOcc,
	group_count(empty, CRow, Empties, Env),
	constrain_empty_counts(CRows, RowOccs, ClmNum, Env).

% ----------------------------------------
% Constraints related to coloured ship pieces

% The following data types will be used in counting ships:

% ship_count_list ==    list(ship_count).       % ShCntL, ShCnts
%                       % A list of ship counts in the rows/columns of the
%			% board, for a fixed ship length and colour.
% ship_count_lists ==   list(ship_count_list)   % ShCntLs
%			A list of ship_count_list-s for all ship lengths
%			(for a fixed colour).
%			Can be thought of as a matrix (not necessarily
%                       rectangular), with the first row containing 
%			counts for ships of length one, the second 
%			row for ships of length two, etc.
% 
% ship_dir --->		/*horizontal:*/ horiz ; /*vertical:*/ vert.

% constrain_colours(ClrDescrs, Clr, CBd, CBdT, Env): The coded board
% CBd/CBdT is constrained appropriately for each element of ClrDescrs, where
% the first element of this list corresponds to colour Clr.
% :- pred constrain_colours(colours_descr::in, colour::in,
%                           coded_board::in, coded_board::in, env::in).
constrain_colours([], _, _, _, _).
constrain_colours([c([Ones|ShCnts],RowOccs,ClmOccs,KnownPs)|ClrDescr],
		 Clr, CBd, CBdT, Env) :-
	constrain_known_positions(KnownPs, Clr, CBd, Env),
	init_ship_counts(ShCntLs0, Env),
	constrain_row_occs(RowOccs, CBd, Clr, horiz, ShCntLs0, ShCntLs1, Env),
	% Because Ones are counted in both horizontal and vertical
	% directions, we extract their ship_count_list, RowOnes, and
	% re-initialise it to empty. 	
	ShCntLs1 = [RowOnes|ShCntLs2], ShCntLs3 = [[]|ShCntLs2],
	constrain_row_occs(ClmOccs, CBdT, Clr, vert, ShCntLs3, ShCntLs4, Env),
	ShCntLs4 = [ClmOnes|ShCntLs],
	% Now ShCntLs contains a ship_count_list for each length
	% 2,3.... Each ship_count_list contains the counts of the given
	% length ships in the rows (horizontal ships) and in the columns
	% (vertical ships).
	% RowOnes contains the counts of singleton ships in the rows,
	% ClmOnes in the columns.
	constrain_ones(Ones, RowOnes, ClmOnes, Env),
	constrain_ship_counts(ShCntLs, ShCnts),
	Clr1 is Clr+1,
	constrain_colours(ClrDescr, Clr1, CBd, CBdT, Env).

% constrain_known_positions(KnownPs, Clr, CBd, Env): The coded board CBd
% is constrained to contain ship pieces of colour Clr at positions 
% and of kind prescribed by KnownPs.
% :- pred constrain_known_positions(known_pieces::in, colour::in,
%                                   coded_board::in, env::in).
constrain_known_positions([], _, _, _).
constrain_known_positions([t(RowI,ClmI,Kind)|KnownPs], Clr, CBd, Env) :-
	matrix_elem(RowI, ClmI, CBd, CF),
	constrain_field(Kind-Clr, CF, Env),
	constrain_known_positions(KnownPs, Clr, CBd, Env).

% ShCntLs is a list of empty ship_count_lists of maximal required length,
% as prescribed by the environment Env.
% :- pred init_ship_counts(ship_count_lists::out, env::in).
init_ship_counts(ShCntLs, Env) :-
	environment(ship_length_max, Env, SLenMax),
	list_of(SLenMax, [], ShCntLs).

% constrain_row_occs(Occs, CRows, Clr, SDir, ShCntLs0, ShCntLs, Env):
% Each coded row in CRows is constrained to contain the appropriate count
% of ship pieces of colour Clr, as prescribed by the corresponding element
% of Occs. 
% Furthermore, for each coded row in CRows, ShCntLs0 is extended
% with counts of ships of colour Clr and direction SDir in the given row,
% for lengths (1..ship_length_max), resulting in ShCntLs.
% 
% Optionally, an important redundant constraint is added for each row:
%      ShCntOtherDir+1*ShCnt_1+2*ShCnt_2+..+M*ShCnt_M #= Occ
% where ShCnt_I is the count of ships of length I in the given row, of the
% given colour and direction, ShCntOtherDir is the count of directed ship 
% pieces of the given colour and opposite direction, and Occ is the count
% of the ship pieces of the given colour in the given row (M is the maximal
% ship length). 
%
% :- pred constrain_row_occs(occs::in, coded_board::in, colour::in,
%                            ship_dir::in, 
%                            ship_count_lists::in, ship_count_lists::out,
%                            env::in). 
constrain_row_occs([], [], _, _, ShCntLs, ShCntLs, _).
constrain_row_occs([Occ|Occs], [CRow|CRows], Clr, SDir,
		   ShCntLs0, ShCntLs, Env) :-
	group_count(all(Clr), CRow, Occ, Env),
	environment(ship_length_seq, Env, ShLens),
	count_ship_heads(ShLens, SDir, Clr, CRow, HeadCnts, Env),
	matrix_prepend(HeadCnts, ShCntLs0, ShCntLs1),
	(   constraint_option(count_ships_occs, Env) ->
	    other_dir(SDir, OtherSDir),
	    group_count(directed(OtherSDir,Clr), CRow, CODir, Env),
	    scalar_product([1|ShLens], [CODir|HeadCnts], #=, Occ)
	;   true
	),
	constrain_row_occs(Occs, CRows, Clr, SDir, ShCntLs1, ShCntLs, Env).

% count_ship_heads(ShLens, SDir, Clr, CRow, ShCnts, Env): ShCnts is the list
% of ship counts of direction SDir and colour Clr in the coded row CRow, for
% each length listed in ShLens.
% :- pred count_ship_heads(list(int)::in, ship_dir::in, colour::in,
%			   coded_row::in, ship_counts::out, env::in).
count_ship_heads([], _, _, _, [], _).
count_ship_heads([ShLen|ShLens], SDir, Clr, CRow, [ShCnt|ShCnts], Env) :-
	group_count(head(SDir,ShLen,Clr), CRow, ShCnt, Env),
	count_ship_heads(ShLens, SDir, Clr, CRow, ShCnts, Env).

% other_dir(SDir, OtherSDir): OtherSDir is the other direction then SDir.
% :- pred other_dir(ship_dir::in, ship_dir::out).
other_dir(horiz, vert).
other_dir(vert, horiz).

% The sum of the list of counts RowOnes is constrained to be equal to Ones.
% Optionally, the same is done for ClmOnes.
% :- pred constrain_ones(int::in, ship_count_list::in, ship_count_list::in,
%			 env::in).
constrain_ones(Ones, RowOnes, ClmOnes, Env) :-
	sum(RowOnes, #=, Ones),
	(   constraint_option(count_ones_columns, Env) ->
	    sum(ClmOnes, #=, Ones)
	;   true
	).

% constrain_ship_counts(ShCntLs, ShCnts): for each element of ShCntLs its
% sum is constrained to be equal to the appropriate element of ShCnts.
% :- pred constrain_ship_counts(ship_count_lists::in, ship_counts::in).
constrain_ship_counts([], _).
constrain_ship_counts([ShCntL|ShCntLs], [ShCnt|ShCnts]) :-
	!, sum(ShCntL, #=, ShCnt),
	constrain_ship_counts(ShCntLs, ShCnts).
constrain_ship_counts([ShCntL|ShCntLs], []) :-
	sum(ShCntL, #=, 0),
	constrain_ship_counts(ShCntLs, []).

% ----------------------------------------
% Constraints related to layout

% Here we use the notion of a generalised list:

% :- type genlist(Type) --->   all(Type)
%                            ; [Type|list(Type)]
%                            ; [].

% all(E) represents a list of undefined length, all elements of which are E.

% Neighbouring fields in the coded board CBd are constrained according to
% their orientation.
% :- pred constrain_layout(coded_board::in, env::in).
constrain_layout(CBd, _Env) :-
	empty_field_code(ECode),
	constrain_rows_layout(CBd, all(ECode), ECode).

% constrain_rows_layout(CRows, CRowUp, ECode): For each field in CRows, its
% neighbours are appropriately constrained, assuming the preceding row is
% CRowUp and the code of an empty field is ECode.
% :- pred constrain_rows_layout(coded_board::in, genlist(coded_field)::in,
%				coded_field::in).
constrain_rows_layout([CRow|CRows], CRowUp, ECode) :-
	constrain_row_layout(CRow, ECode, CRowUp, ECode, ECode),
	constrain_rows_layout(CRows, CRow, ECode).
constrain_rows_layout([], CRowLast, ECode) :-
	constrain_row_layout(all(ECode), ECode, CRowLast, ECode, ECode).

% constrain_row_layout(CFs, CFLeft, CFsUp, CFUpLeft, ECode): For each
% field in CFs its neighbours are appropriately constrained, assuming the
% field to its left is CFLeft, the row above it is CFsUp, the field above
% CFLeft is CFUpLeft and the code of an empty field is ECode.
% :- pred constrain_row_layout(genlist(coded_field)::in, coded_field::in,
%			       genlist(coded_field)::in, coded_field::in,
%			       coded_field::in).
constrain_row_layout(CFs, CFLeft, CFsUp, CFUpLeft, ECode) :-
	gen_list_head_tail(CFs, CF, CFs1),
	gen_list_head_tail(CFsUp, CFUp, CFsUp1), !,
	constrain_field_layout(CF, CFLeft, CFUp, CFUpLeft),
	constrain_row_layout(CFs1, CF, CFsUp1, CFUp, ECode).
constrain_row_layout(_, CFLast, _, _, ECode) :-
	coded_field_neighbour(horiz, CFLast, ECode).

% gen_list_head_tail(GList, H, T): the head and tail of the generalised
% list GList are H and T.
% :- pred gen_list_head_tail(genlist(T)::in, T::out, genlist(T)::out).
gen_list_head_tail(all(N), N, all(N)).
gen_list_head_tail([H|T], H, T).

% Appropriate constraints are set up between the four pieces arranged as
% follows:           
%                 UpLeft        Up
%                 Left          Current
% :- pred constrain_field_layout(coded_field::in, coded_field::in,
%				 coded_field::in, coded_field::in).
constrain_field_layout(Current, Left, Up, UpLeft) :-
	coded_field_neighbour(diag, Current, UpLeft),
	coded_field_neighbour(diag, Up, Left),
	coded_field_neighbour(horiz, Left, Current),
	coded_field_neighbour(vert, Up, Current).


% ----------------------------------------
% Converting a coded board to output format

% coded_board_to_board(CRows, Rows, Env): The output format of CRows is Rows.
% :- pred coded_board_to_board(coded_board::in, board::out, env::in).
coded_board_to_board([], [], _).
coded_board_to_board([CRow|CRows], [Row|Rows], Env) :-
	coded_row_to_row(CRow, Row, Env),
	coded_board_to_board(CRows, Rows, Env).

% coded_row_to_row(CFields, Fields, Env): The output format of CFields 
% is Fields. 
% :- pred coded_row_to_row(coded_row::in, row::out, env::in).
coded_row_to_row([], [], _).
coded_row_to_row([CF|CFs], [F|Fs], Env) :-
	coded_field_to_field(CF, F, Env),
	coded_row_to_row(CFs, Fs, Env).


