% File torp_bed.pl 
% Testbed for the torpedo puzzle
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(torpedo_testbed,
          [time_torpedo/2, time_torpedo/1, problem/2,
	   all_tests/1, simple_tests/1, hard_tests/1, long_tests/1,
% re-exporting from torp_utils:	   
	   show_board/2, show_table/2, show_empty_table/2,
	   table_to_problem/3, board_to_problem/3,
% re-exporting from torp:
	   torpedo/2, torpedo/3
	   ]).

% :- import torpedo/2. % in version 1 compatibility mode
% torpedo(TDescr, Bd): Bd is a solution of of the torpedo problem TDescr.
% :- pred torpedo(torp_descr::in, board::out).

% :- import torpedo/3. 
% torpedo(TDescr, Bd, Opts): Bd is a solution of of the torpedo
% problem TDescr, found using options Opts.
% :- pred torpedo(torp_descr::in, board::out, options::in).

%!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
% file torp should start with:         
% :- module(torpedo, [torpedo/...]).
%!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
:- use_module(torp).

:- use_module(torp_utils).
:- use_module(utils).

:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- use_module(library(timeout)).
:- use_module(library(system)).

%-------------------- The data type for stating the torpedo problem

%         Type               Definition                        Variable names
% ----------------------------------------------------------------------------
% :- type torp_descr --->    descr(known_empties, colours_descr).  % TDescr
% :- type known_empties ==   list(empty_field).                % KnownEs, KEs
% :- type empty_field --->   row_index-column_index.           % KnownE, KE
% :- type colours_descr ==   list(colour_descr).               % ClrDescrs, CDs
% :- type colour_descr --->  c(ship_counts, 
%                              row_occs, column_occs,
%                              known_pieces).                  % ClrDescr, CD
% :- type ship_counts ==     list(ship_count)                  % ShCnts, SCs
%			       /* for compatibility only: */
%	                   | pieces(/*ones:*/ ship_count,
%                                   /*twos:*/ ship_count,
%                                   /*threes:*/ ship_count,
%                                   /*fours:*/ ship_count).
% :- type row_occs ==        occs.                             % RowOccs, ROs
% :- type column_occs ==     occs.                             % ClmOccs, COs
% :- type occs ==            list(occ).                        % Occs, Os
% :- type known_pieces ==    list(known_piece).                % KnownPs, KPs
% :- type known_piece --->   t(row_index, column_index, kind). % KnownP, KP
%			     % the colour is implicit.             
% :- type row_index ==       int. % 1,2,...                    % RowI
%			     % (rows numbered from 1) 
%                                     % maximal row index:     % RowNum
% :- type column_index ==    int. % 1,2,...                    % ClmI
%			     % (columns numbered from 1) 
%                                     % maximal column index:  % ClmNum
% :- type occ ==             int.                              % Occ, RO, CO
% :- type ship_count ==      int.                              % ShCnt
% :- type kind --->              u
%                          ; l ; m ; r
%                          ;     d 
%                          ;     o  .                          % Kind

%-------------------- The data type for returning the solution

% :- type board  ==          list(row).                        % Bd, Rows
% :- type row ==             list(field).                      % Row, Fs
% :- type field --->         =  /* empty */                   
%                          ; kind-colour.  /* piece */         % Field, F
% :- type colour ==          int. % 1,2,...
%			     % (colours numbered from 1)       % Clr
%                                                  % maximal:  % ClrMax


%-------------------- Sample problems

% :- type problem_descr == any.                                % PDescr
%                          % ( torp_descr | atom | problem_descr - modif)

% named_problem(Name, PDescr): Name names problem PDescr.
% :- pred named_problem(atom::in, problem_descr::out)
% The unique solution (type 'board') to the problem 'simple' below is:
% [[ = ,l-1,r-1, = , = ],
%  [ = , = , = , = ,o-2],
%  [u-2, = , = , = , = ],
%  [d-2, = , = ,o-1, = ]]
named_problem(simple,          
    descr([1-1],
         [c([1,1], [2,0,0,1], [0,1,1,1,0], [t(1,3,r)]),
          c([1,1], [0,1,1,1], [2,0,0,0,1], [])]
        )).
% The problem 'fules' is the one distributed at the course. Single solution.
named_problem(fules, 
    descr([2-5,9-1,9-10,12-10],
         [c([4,3,2,1],
                [0,3,1,0,4,0,3,0,4,1,3,0,1],
                [0,2,2,2,2,2,2,1,1,1,0,4,1],
                [t(2,3,r), t(5,4,l), t(7,8,o), t(10,12,o)]),
          c([4,3,2,1],
                [1,2,2,1,2,2,1,1,0,0,2,2,4],
                [3,4,0,0,1,2,4,0,0,4,0,0,2],
                [t(6,2,m), t(8,6,o)])]
        )).
% The one below is a very dense 3-colour problem (density = 68/169 ~ 40%).
% Has a single solution.
named_problem(dense,
    descr([1-11,3-2],
	  [c([3,3,3,1],
		  [5,0,3,1,2,0,2,2,3,0,0,0,4],
		  [5,0,3,0,2,0,2,2,2,1,0,2,3],
		  [t(13,1,o)]),
	   c([3,2,3,2],
		  [2,0,5,1,1,0,6,0,3,1,3,1,1],
		  [1,0,4,0,4,0,3,3,2,5,0,1,1],
		  [t(1,3,o),t(7,8,m)]),
	   c([3,3,3,1],
		  [2,0,1,1,5,0,1,1,1,1,5,1,3],
		  [3,0,3,0,4,0,2,2,1,0,2,2,3],
		  [t(11,5,m),t(13,11,o)])]
	     )).
% This is the problem 'fules', but with no known pieces. Has 100 solutions.
named_problem(fules_clean, fules-clean).
% Problem 'fules_clean' but with most of fields in the diagonals known
% to be empty. It has two solutions.
named_problem(fules2,
    fules_clean-
        empties([1-1,3-3,4-4,6-6,7-7,8-8,9-9,10-10,11-11,12-12,13-13,
                 1-13,2-12,3-11,4-10,5-9,6-8,8-6,9-5,10-4,11-3,12-2])).
% A variant of 'fules2', with slightly less fields known to be empty. 
% It has 6 solutions.
named_problem(fules2a,
    fules_clean-
        empties([1-1,3-3,4-4,6-6,7-7,8-8,9-9,10-10,11-11,12-12,13-13,
		 1-13,2-12,3-11,4-10,5-9,6-8,8-6,9-5])).
% Problem 'fules_clean' but with most of fields in the NW-SE diagonal known
% to be empty. It has 10 solutions.
named_problem(fules3,
    fules_clean-
        empties([1-1,3-3,4-4,6-6,7-7,8-8,9-9,10-10,11-11,12-12,13-13])).
named_problem(test1, descr([],
         [c([2], [1,0,1], [1,0,1], []),
	  c([2], [1,0,1], [1,0,1], [])])).
named_problem(test2, descr([],
         [c([0,2], [2,0,2], [1,1,0,1,1], []),
	  c([0,2], [2,0,2], [1,1,0,1,1], [])])).
named_problem(test3, descr([],
         [c([4,4], [3,0,3,0,3,0,3], [2,1,1,2,0,2,1,1,2], []),
	  c([4,4], [3,0,3,0,3,0,3], [2,1,1,2,0,2,1,1,2], [])])).

%-------------------- Problem description processing

% :- export problem/2.
% problem(PDescr, TDescr):  PDescr names TDescr.
% :- pred problem(problem_descr::in, torp_descr::out).
problem(PDescr0-Mod, TDescr) :-
        !, problem(PDescr0, TDescr0),
        modify_problem(Mod, TDescr0, TDescr).
problem(PDescr0, TDescr) :-
        named_problem(PDescr0, PDescr1), !,
        problem(PDescr1, TDescr).
problem(TDescr, TDescr).

% :- type modif --->         empties(known_empties)
%                          ; pieces(colour, known_pieces)
%                          ; clean
%                          ; collapse.

% modify_problem(Mod, TDescr0, TDescr): the torpedo problem TDescr0, 
%                   transformed according to modification Mod, is TDescr.
% modify_problem(empties(KEs),...):
%				   Set the known_empties list to KEs.
% modify_problem(pieces(Clr,KPs),...): 
%				   Set the known_pieces list
%                                  for colour Clr to KPs.
% modify_problem(clean,...):       Set the known_empties list and all
%                                  known_pieces lists to [].
% modify_problem(collapse,...):    Collapse all colours, i.e. create a problem 
%                                  with a single colour, with the cues for the
%                                  colours combined appropriately.
% :- pred modify_problem(modif::in, torp_descr::in, torp_descr::out).

modify_problem(empties(KEs),    descr(_,CDs), descr(KEs,CDs)).
modify_problem(pieces(Clr,KPs), descr(KEs,CDs0), descr(KEs,CDs)) :-
        replace_nth(Clr, CDs0, c(SCs,ROs,COs,_), CDs, c(SCs,ROs,COs,KPs)).
modify_problem(clean,           descr(_,CDs0),
	                        descr([],CDs)) :-
        findall(c(SCs,ROs,COs,[]), member(c(SCs,ROs,COs,_), CDs0), CDs).
modify_problem(collapse,        descr(KEs,CDs), descr(KEs,[CD_Collapsed])) :-
        collapsed_colours_descr(CDs, CD_Collapsed).

%-------------------- Solving the problem and displaying statistics

% Display options:
% :- type options ==         list(option).                     % Opts
% :- type option --->        skip(skip_what)                   
%                          ; stop_at(int)
%			   ; {user-defined option}.            % Opt
% :- type skip_what --->     board ; stats ; all.

% skip(board):               do not show the solution itself,
% skip(stats):               do not show statistics for the solution,
% skip(all):                 equivalent to skip(board) and skip(stats),
% stop_at(Sols):             stop after Sols solutions found.

% time_torpedo(PDescr): Solve the torpedo problem PDescr and display 
% statistics with default options.
% :- export time_torpedo/1.
% :- pred time_torpedo(problem_descr::in).
time_torpedo(PDescr) :-
        time_torpedo(PDescr, []).

% time_torpedo(PDescr, Opts): Solve the torpedo problem PDescr and display
% statistics according to options Opts.
% :- export time_torpedo/2.

% :- pred time_torpedo(problem_descr::in, options::in).
time_torpedo(PDescr, Opts0) :-
	expand_options(Opts0, Opts1),
	v1_compatibility(Opts1, Opts),
	format(user_error,
	       '~n-------------------------~nProblem: ~w, Options: ~w~n~n', 
	       [PDescr,Opts]),
        problem(PDescr, TDescr0),
	adjust_torp_descr_for_v1(TDescr0, TDescr, Opts),
	init_stats,
	timeout_value(Opts, TimeOut),
	protected_torpedo(TDescr, TimeOut, Opts, Res),
        (   Res = success(Bd) ->
	    show_solution(PDescr, TDescr, Bd, Opts, SolNum),
            member(stop_at(SolNum), Opts), !,
	    Sols = 'not all'
	;   Res = failure ->
	    show_failure(PDescr, 'Failure', Opts),
	    Sols = 'all'
        ;   show_failure(PDescr, Res, Opts),
	    Sols = Res
        ),
	show_totals(PDescr, Sols, Opts),
	store_or_check_solutions(PDescr, Opts).

% Time_msec is the allowed runtime in msecs, according  to Opts.
% :- pred timeout_value(options::in, int::out).
timeout_value(Opts, Time_msec) :-
	member(timeout(T), Opts), !,
	Time_msec is T*1000.
timeout_value(_, 240000).  % 4 minutes

% :- type result ---> exception(any) ; success(board) ; time_out ; failure.
% The result of running torpedo problem TDescr with TimeOut and Opts is Res.
% :- pred protected_torpedo(torp_descr::in, int::in, options::in, result::out).
protected_torpedo(TDescr, TimeOut, Opts, Res) :-
	on_exception(Exc, 
		     time_out(call_torpedo(TDescr, Bd, Opts), TimeOut, Res0),
		     Res0 = exception(Exc)
		    ),
	(   Res0 = success -> Res = success(Bd)
	;   Res = Res0, !
	).
protected_torpedo(_, _, _, failure).

% Bd is a solution of of the torpedo problem TDescr, found using options Opts.
% If Opts includes v1 uses torpedo/2, otherwise torpedo/3.
% :- pred call_torpedo(torp_descr::in, board::out, options::in).
call_torpedo(TDescr, Bd, Opts) :-
	member(v1, Opts), !,
	torpedo(TDescr, Bd).
call_torpedo(TDescr, Bd, Opts) :-
	torpedo(TDescr, Bd, Opts).

% init_stats: initialise statistics.
% :- pred init_stats.
init_stats :-
        fd_statistics(backtracks, _), statistics(runtime, [T0,_]),
        bb_put(btracks_total, 0), bb_put(sol_num, 0),
	bb_put(sol_list, []),
        bb_put(time, T0), bb_put(time_total, 0).

% Show statistics for problem PDescr/TDescr, solution Bd, solution number S1.
% Display statistics according to Opts.
% :- pred show_solution(problem_descr::in,
%                       torp_descr::in, board::in, options::in, int::out).
show_solution(PDescr, TDescr, Bd, Opts, SolNum) :-
        get_time_btracks(T, B),
	store_solution(Bd, Opts, SolNum),
        (   member(skip(all), Opts) -> true
        ;   \+ member(skip(board), Opts), show_board(TDescr, Bd), fail
        ;   \+ member(skip(stats), Opts), 
            format(user_error,
		   'Problem: ~w~t~25|Solution:~t~d ~13+~50|~t~3d~7+ sec,~t~D~13+ btracks.~n',
		   [PDescr,SolNum,T,B]),
            fail
        ;   true
        ),
        statistics(runtime, [T2,_]),
        bb_put(time, T2).

% T is the time and and B is the number of backtracks done since the
% last solution.
% :- pred get_time_btracks(int::out, int::out).
get_time_btracks(T, B) :-
        statistics(runtime, [T1,_]),
        bb_get(time, T0),
        T is T1-T0,
        fd_statistics(backtracks, B),
        bb_get(btracks_total, B0), B1 is B+B0,
        bb_put(btracks_total, B1),
	bb_get(time_total, TT0), TT1 is TT0+T,
	bb_put(time_total, TT1).

% Increment the sol_num counter, return in SolNum the solution number of
% the current solution. Store the Bd, if requested in Opts.
% :- pred store_solution(board::in, options::in, int::out).
store_solution(Bd, Opts, SolNum) :-
        bb_get(sol_num, S),
        SolNum is S+1, bb_put(sol_num, SolNum),
	(   member(sols(_), Opts) ->
	    bb_get(sol_list, Bds),
	    bb_put(sol_list, [Bd|Bds])
	;   true
	).

% Store or check the solution list for problem PDescr, as requested in Opts.
% :- pred store_or_check_solutions(problem_descr::in, options::in).
store_or_check_solutions(PDescr, Opts) :-
	member(sols(save), Opts), !,
	bb_get(sol_list, Bds0),
	sort(Bds0, Bds),
	datime(Date),
	open('torp_sols.pl', append, S),
	write(S, '% '),
	format_date(S, Date), nl(S),
	format(S, 'torp_solutions(~q,~q).~n', [PDescr,Bds]),
	close(S).
store_or_check_solutions(PDescr, Opts) :-
	member(sols(check), Opts), !,
	bb_update(sol_list, Bds0, Bds0),
	sort(Bds0, Bds),
%	prolog:'$lowdbg'(_, 1),
	ensure_loaded(torp_sols),
	(   current_predicate(_, torp_solutions(_,_)),
	    torp_solutions(PDescr, BdsStored) -> 
	    (	Bds = BdsStored -> true
	    ;	format(user_error,
		       '{ERROR: problem ~w, solution list differs from stored}~n~q~n',
		       [PDescr,Bds])
	    )
	;   format(user_error, '{Warning: Problem ~w, stored solution not found}~n',
		   [PDescr])
	).
store_or_check_solutions(_, _).

% Show statistics for failure of problem PDescr caused by Cause,
% according to Opts.
% :- pred show_failure(problem_descr::in, any::in, options::in).
show_failure(PDescr, Cause, Opts) :-
        get_time_btracks(T, B),
        (   member(skip(stats), Opts) -> true
        ;   member(skip(all), Opts) -> true
        ;   format(user_error,
               '~nProblem: ~w~t~25|~w: ~50|~t~3d~7+ sec,~t~D~13+ btracks.~n',
		   [PDescr,Cause,T,B])
        ).

% Show statistics totals for problem PDescr.
% Txt indicates whether all or some solutions were found.
% :- pred show_totals(problem_descr::in, atom::in, options::in).
show_totals(PDescr, Txt, Opts) :-
        bb_get(btracks_total, BTot),             
        bb_get(time_total, TTot),             
        bb_get(sol_num, SolNum),
        format(user_error,
	       '~nProblem: ~w~t~25|Solutions:~t~d~13+ (~w), ~50|~t~3d~7+ sec,~t~D~13+ btracks.~n',
               [PDescr,SolNum,Txt,TTot,BTot]),
	(   member(save_stats, Opts) ->
	    datime(Date),
	    remove_testbed_options(Opts, Opts1),
	    unexpand_options(Opts1, Opts2),
	    open('torp_stat.pl', append, S),
	    format(S, 'stat(~q,~q,~t~47|', [PDescr,Opts2]),
	    (	line_position(S, 47) -> true
	    ;	nl(S), tab(S, 47)
	    ),
	    format_date(S, Date),
	    format(S, ',~|~t~3d~10+,~t~d~8+).~n', [TTot,BTot]),
	    close(S)
	;   true
	).

% remove_testbed_options(Opts, Opts1): Removing the testbed-options from
% Opts gives Opts1.
% :- pred remove_testbed_options(options::in, options::out).
remove_testbed_options([], []).
remove_testbed_options([Opt|Opts], Opts1) :-
	testbed_option(Opt), !,
	remove_testbed_options(Opts, Opts1).
remove_testbed_options([Opt|Opts], [Opt|Opts1]) :-
	remove_testbed_options(Opts, Opts1).

% testbed_option(Opt): Opt is an option of the torpedo testbed.
% :- pred testbed_option(option::out).
testbed_option(skip(_)).
testbed_option(stop_at(_)).
testbed_option(v1).
testbed_option(sols(_)).
testbed_option(save_stats).
testbed_option(timeout(_)).

format_date(S, datime(Year,Month,Day,Hour,Min,_Sec)) :-
	Year1 is Year mod 100,
	format(S, '~|~`0t~d~2+~`0t~d~2+~`0t~d~2+:~`0t~d~3+~`0t~d~2+',
	       [Year1,Month,Day,Hour,Min]).

%-------------------- Sample test cases

all_tests(Opts) :-
	simple_tests(Opts), hard_tests(Opts), long_tests(Opts).

simple_tests(Opts) :-
	time_torpedo(simple,          Opts+stop_at(2)),
	time_torpedo(fules,           Opts+stop_at(2)),
	time_torpedo(simple-collapse, Opts+stop_at(2)),
	time_torpedo(dense,           Opts+stop_at(3)),
	time_torpedo(fules-collapse,  Opts+stop_at(3)).
				          	      
hard_tests(Opts) :-		          	      
	time_torpedo(fules2,          Opts+stop_at(3)),
	time_torpedo(dense-clean,     Opts+stop_at(3)),
	time_torpedo(dense-collapse,  Opts+stop_at(2)).

long_tests(Opts) :-
	time_torpedo(fules2a,         Opts+stop_at(11)+timeout(1000)),
	time_torpedo(fules3,          Opts+stop_at(11)+timeout(1000)),
	time_torpedo(fules_clean,     Opts+stop_at(101)+timeout(3600)),
	true.


%-------------------- Expanding named options

expand_options(Name, Opts) :-
	named_option(Name, Opts0), !, expand_options(Opts0, Opts).
expand_options(Opts+Opt, EOpts) :-
	!, expand_options(Opts, EOpts0), remove_opt(Opt, EOpts0, EOpts1),
	EOpts = [Opt|EOpts1].
expand_options(Opts-Opt, EOpts) :-
	!, expand_options(Opts, EOpts0), remove_opt(Opt, EOpts0, EOpts).
expand_options(Opts, Opts).

% Opts0 is an expanded form of an option name Opts, or equal to Opts.
unexpand_options(Opts0, Opts) :-
	named_option(Name, Opts1), expand_options(Opts1, Opts2),
	Opts2 = Opts0, !, Opts = Name.
unexpand_options(Opts, Opts).

% remove_opt(Opt, Opts0, Opts): removing Opt from Opts0 yields Opts.
remove_opt(Opt, Opts0, Opts) :-
	select(Opt, Opts0, Opts), !.
remove_opt(Opt0, Opts0, Opts) :-
	functor(Opt0, F, A), F \= +, F \= -, functor(MGT,F, A),
	select(MGT, Opts0, Opts), !.
remove_opt(_Opt, Opts, Opts).


%-------------------- Some named option-sets

% named_option(Name, Opts): Name is a name for options Opts
% ------------
% `base' is the simplest setup:
% diagonal constraints implemented via reification,
% relation/3 not expanded to an indexical, no filtering,
% plain labeling, no redundant constraints.
named_option(base,
	      [diag(reif),relation(clause),filter([off]),
	       label(plain),+all_redundant_off]).
named_option(base_cs, base+ +(count_ships_occs)).
named_option(base_cso, base_cs+ +(count_ones_columns)).
% `redund' is 'base' with the redundant constraints enabled.
named_option(redund, % == base_cso+ +(count_empties)).
	      [diag(reif),relation(clause),filter([off]),label(plain)]).
% variants using more sophisticated labeling:
named_option(max_dual,     redund+label(max_dual)).
named_option(ships,        redund+label(ships)).
% variants using filtering:
named_option(filter_repet, ships+filter([repetitive])).
named_option(filter,       ships+filter([on])).
% expanding relation/3 to an indexical:
named_option(rel_ind, filter+relation(indexical)).
% variants using indexicals for the diagonal constraints:
named_option(diag_arith, rel_ind+diag(ind_arith)).
named_option(diag_cond, rel_ind+diag(ind_cond)).

%-------------------- Version 1 compatibility

v1_compatibility(Opts0, Opts) :-
	current_predicate(_, torpedo:torpedo(_,_,_)), !, Opts = Opts0.
v1_compatibility(Opts0, [v1|Opts0]).
