% File torp_utils.pl 
% Utilities for the torpedo puzzle (displaying the board, etc.)
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(torpedo_utilities,
          [show_board/2, show_table/2, show_empty_table/2,
	   table_to_problem/3, board_to_problem/3,
	   collapsed_colours_descr/2,
	   adjust_torp_descr_for_v1/3]).

:- use_module(utils).
:- use_module(library(lists)).

%-------------------- Displaying the board

% A sample display:         Explanation:

%       1 2 3 4 5           <--- column numbers (mod 10)
% 
%       0 1 1 1 0           <--- no. of occurrences of colour 1 in columns
% 
% 1  2  = * r : :  0        <---
% 2  0  : : : : #  1        <---
% 3  0  # : : : :  1        <---  board proper (fields)
% 4  1  # : : * :  1        <---
% 
%       2 0 0 0 1           <--- no. of occurrences of colour 2 in columns
%
% ^----------------------------- row numbers (mod 10)
%    ^-------------------------- no. of occurrences of colour 1 in rows
%       ^-^-^-^-^--------------- board proper (fields)
%                  ^------------ no. of occurrences of colour 2 in rows

% Pieces legend:
%
%   =    			known to be empty
%   :    			deduced to be empty
%   <lower case letter> 	known to be of the kind given and of colour 1
%				(letter is one of u,d,m,l,r,o)
%   <upper case letter>		known to be of the kind given and of colour 2
%				(letter is one of U,D,M,L,R,O)
%   <lower case letter><digit>  known to be of the kind given and of colour
%				<digit>, <digit> >= 3
%   *           		deduced to be of colour 1
%   #           		deduced to be of colour 2
%   <digit>    			deduced to be of colour <digit>, <digit> >= 3

% The "board proper" part of the output carries all information needed to
% reproduce the original problem.

% --------Limits of the display

% A board is displayed properly iff all 'colour's in it are =< 9, and
% the 'occ's for the first two colours are =< 99.
% If there are column occs >= 10, these are displayed in two lines,
% vertically (see the solution of dense-collapse among the examples at the
% end of this file, for an example of this).

% Type style specifies the way the board, or a row of the board, is shown:
% :- type style --->
%     nums            % display column labels 1, 2, ...
%   ; tens            % display the first digit of two digit
%                     % occurrence numbers (if any)
%   ; units           % display the last digit of occ. numbers
%   ; base(int,torp_descr,sub_style).
%                     % display a base row for a given problem description, 
%                     % with a given row number, and using the given sub-style.

% :- type sub_style --->
%     occs(occs,occs) % show the occ. numbers on the left and on the
%                     % right of the board
%   ; string(atom).
%                     % display the row as a string preceded by the atom
             
% :- export show_board/2.
% Show the board Bd for problem TDescr.
% :- pred show_board(torp_descr::in, board::in).
show_board(TDescr, Bd) :-
        TDescr = descr(_,ClrDescrs),
        first_two_colour_occs(ClrDescrs, RowOccs1, ClmOccs1,
			                 RowOccs2, ClmOccs2), 
        nl,
        show_row(ClmOccs1, nums),
        nl,
        show_row(ClmOccs1, tens),
        show_row(ClmOccs1, units),
        nl,
        show_rows(Bd, base(1,TDescr,occs(RowOccs1,RowOccs2))),
        nl,
        show_row(ClmOccs2, tens),
        show_row(ClmOccs2, units),
        nl.

% The row and column occurrences for the first two colours in ClrDescrs are
% RowOccs1, ClmOccs1 and RowOccs2, ClmOccs2, respectively. The latter two 
% are [] for single colour problems.
% :- pred first_two_colour_occs(colours_descr::in,
%         row_occs::out, column_occs::out, row_occs::out, column_occs::out).
first_two_colour_occs(ClrDescrs, RowOccs1, ClmOccs1, RowOccs2, ClmOccs2) :-
        ClrDescrs = [c(_,RowOccs1,ClmOccs1,_)|ClrDescrs1],
        (   ClrDescrs1 = [c(_,RowOccs2,ClmOccs2,_)|_] -> true
        ;   RowOccs2 = [], ClmOccs2 = []
        ).

% show_rows(Rows, Style): Show the rows Rows using style Style.
% :- pred show_rows(board::in, style::in).
show_rows([Row|Rows], Style) :-
        show_row(Row, Style),
        next_style(Style, Style1),
        show_rows(Rows, Style1).
show_rows([], _).

% next_style(St0, St): Moving to the next row in style St0 gives St.
% :- pred next_style(style::in, style::out).
next_style(base(RowI,TDescr, SStyle), base(RowI1,TDescr,SStyle1)) :- !,
	RowI1 is RowI+1,
	next_sub_style(SStyle, SStyle1).
next_style(St, St).

% next_sub_style(SSt0, SSt): Moving to the next row in  SSt0 gives SSt.
% :- pred next_sub_style(sub_style::in, sub_style::out).
next_sub_style(occs([_|RowOccs1],[_|RowOccs2]), SStyle) :- !,
	SStyle = occs(RowOccs1,RowOccs2).
next_sub_style(occs([_|RowOccs1],[]), occs(RowOccs1,[])).
next_sub_style(string(_), string(',\n ')).

% Show row Row in style Style.
% :- pred show_row(row::in, style::in).
show_row(Row, Style) :-
        show_row_header(Style, Row), !,
        show_fields(Row, Style, 1),
        show_row_trailer(Style).
show_row(_R, _Style).

% show_row_header(St, Row): Show the header for row Row, style St.
% Fail if the row should not be shown.
% :- pred show_row_header(style::in, row::in)
show_row_header(base(RowI,_,SubStyle), _) :- !,
	show_base_row_header(SubStyle, RowI).
show_row_header(_, []) :- !, fail.
show_row_header(tens, Row) :-
        \+ (member(E, Row), E >= 10), !, fail.
show_row_header(_, _) :-
        format('~t~4| ', []).

% show_base_row_header(SSt, RowI): Show the header for a base row, 
% sub-style SSt and row number RowI.
% :- pred show_base_row_header(sub_style::in, int::in)
show_base_row_header(string(Prefix), _) :-
        write(Prefix), write('"'). %"
show_base_row_header(occs([Occ1|_],_), RowI) :-
        RowI1 is RowI mod 10,
        format('~w~t~w~4| ', [RowI1,Occ1]).

% show_row_trailer(St): Show the row trailer in style St.
% :- pred show_row_trailer(style::in)
show_row_trailer(base(_,_,SubStyle)) :- !,
	(   SubStyle = string(_) ->
	    write('"')		%"
	;   SubStyle = occs(_,[Occ2|_]) ->
	    format('~|~t~w~3+~n', [Occ2])
	;   nl
	).
show_row_trailer(_) :-
        nl.

% Show fields Fs0 in style Style column ClmI.
% :- pred show_fields(row::in, style::in, int::in).
show_fields(Fs0, Style, ClmI) :-
        decomp_fields(Style, Fs0, ClmI, F, Fs), !, 
        show_field(F),
        ClmI1 is ClmI+1,
        show_fields(Fs, Style, ClmI1).
show_fields(_, _, _).

% type out_field == any.  % the output format of a field.
%                % ( single_letter_atom | digit | kind-colour)
% decomp_fields(Style, Fs0, ClmI, F, Fs): Split a fields description Fs0, to
% be shown in style Style and at column ClmI, to an (output format) field F
% and rest Fs.
% :- pred decomp_fields(style::in, row::in, int::in, out_field::out, row::out).
decomp_fields(tens, [F0|Fs], _, F, Fs) :-
        !, F1 is F0 // 10,
        (   F1 > 0 -> F = F1
        ;   F = ' '
        ).
decomp_fields(units, [F0|Fs], _, F, Fs) :-
        !, F is F0 mod 10.
decomp_fields(nums, [_|Fs], ClmI, F, Fs) :-
        !, F is ClmI mod 10.
decomp_fields(base(RowI,TDescr,_), [F0|Fs], ClmI, F, Fs) :-
	!, convert_field(F0, RowI, ClmI, TDescr, F).
decomp_fields(_, [F|Fs], _, F, Fs).

% convert_field(F0, RowI, ClmI, TDescr, F): the output format of a field F0
% at (RowI,ClmI), for problem TDescr is F.
% :- pred convert_field(field::in, int::in, int::in, torp_descr::in
%                       out_field::out).
convert_field(V, _, _, _, ' ') :- var(V), !.
convert_field(F0, RowI, ClmI, descr(_,ClrDescrs), F) :-  F0 = _-Clr, !,
        nth(Clr, ClrDescrs, c(_,_,_,KnownPs)),
        (   memberchk(t(RowI,ClmI,_), KnownPs) -> F = F0     % known
        ;   convert_colour(Clr, F)                           % deduced
        ).
convert_field(=, RowI, ClmI, descr(KnownEs,_), F) :- !,
        (   memberchk(RowI-ClmI, KnownEs) -> F = '='         % known
        ;   F = ':'                                          % deduced
        ).
convert_field(F, _, _, _, F).

% show_field(F): display field F.
% :- type show_field(field1::in)
show_field(Kind-Clr) :- !,
	show_colour_field(Clr, Kind).
show_field(F) :-
	format('~|~t~w~2+', [F]).

% show_colour_field(Clr, Kind): show a field of kind Kind, colour Clr.
% :- pred show_colour_field(colour::in, kind::in).
show_colour_field(1, Kind) :-
        !, tab(1), write(Kind).
show_colour_field(2, Kind) :-
        !, atom_chars(Kind, [Code]),
        UpCode is Code-0'a+0'A,
        tab(1), put(UpCode).
show_colour_field(Clr, Kind) :-
        write(Kind), write(Clr).

% convert_colour(Colour, Char): Char is used for displaying (deduced)
% fields of colour Colour.
% :- pred convert_colour(colour::in, atomic::out).
convert_colour(1, *) :- !.
convert_colour(2, #) :- !.
convert_colour(Clr, Clr).

%-------------------- Tables: boards displayed in re-readable format

% :- type table == list(table_row).
% :- type table_row == list(char_code).
% :- type char_code == int.
% The table format of a board is a list of strings, where each string
% corresponds to a row.  Each field in a row is represented by two
% characters in the string, exactly as in the proper board part of
% the output produced by show_board/2 (see the Pieces legend above the
% definition of show_board/2, note that single character fields are
% prefixed by a space).

% :- export show_table/2.
% Output a prolog term, which corresponds to the table format of the
% board Bd for problem TDescr. This can then be edited and converted to a
% problem using table_to_problem/3, see below.
% :- pred show_table(torp_descr::in, board::in).
show_table(TDescr, Bd) :-
        show_rows(Bd, base(1,TDescr,string('['))),
        write(']'), nl.

% :- export show_empty_table/2.
% Output a prolog term, which corresponds to the table format of an empty
% board with RowI rows and ClmI columns. This can then be edited and
% converted to a problem using table_to_problem/3, see below.
% :- pred show_empty_table(int::in, int::in).
show_empty_table(RowI, ClmI) :-
        matrix_of(RowI, ClmI, :, Bd),
	show_rows(Bd, base(1,descr([],[]),string('['))),
        write(']'), nl. 

%-------------------- Converting a table to a board

% The board1 data type is exactly the same as the board data type,
% but distinguishing between known and deduced fields. This carries all
% information present in a `table' data structure, thus allowing the
% reconstruction of a problem from a board1 instance.

% :- type board1  == list(row1).
% :- type row1 ==    list(field1).
% :- field1 --->     =                % known empty
%                  ; :                % deduced empty
%                  ; kind1-colour.
% :- type kind1 =        u
%                  ; l ; m ; r        % known to be of the given kind
%                  ;     d
%                  ;     o
%                  ;     x.           % deduced to be non-empty


% table_to_board(Tbl, Bd): The board1 corresponding to table Tbl is Bd.
% :- pred table_to_board(table::in, board1::out).
table_to_board([], []).
table_to_board([TR|TRs], [Row|Rows]) :-
        table_row_to_board_row(TR, Row),
        table_to_board(TRs, Rows).

% table_row_to_board_row(Ts, Es): Es is a board row corresponding to the
% table row Ts.
% :- pred table_row_to_board_row(table_row::in, row1::out).
table_row_to_board_row([], []).
table_row_to_board_row([T1,T2|Ts], [E|Es]) :-
        table_elem_to_board_elem(T1, T2, E),
        table_row_to_board_row(Ts, Es).

% table_elem_to_board_elem(T1, T2, E)): E is a board field corresponding
% to the table characters T1 and T2.
% :- pred table_elem_to_board_elem(char_code::in, char_code::in, field1::out).
table_elem_to_board_elem(0' , T, E) :- !,
        (   table_abbrev(T, T1, T2) ->
            table_elem_to_board_elem(T1, T2, E)
        ;   atom_chars(E, [T])
        ).
table_elem_to_board_elem(F, Clr, E) :-
        digit_code(Clr), !,
        Clr1 is Clr-0'0, atom_chars(F1, [F]),
        E = F1-Clr1.

% table_abbrev(C, C1, C2): the single character C is an abbreviation of
% the two characters C1 and C2.
% :- pred table_abbrev(char_code::in, char_code::out, char_code::out).
table_abbrev(0'*, 0'x, 0'1).
table_abbrev(0'#, 0'x, 0'2).
table_abbrev(D, 0'x, D) :-
        digit_code(D).
table_abbrev(LC, LC, 0'1) :-
        lc_code(LC).
table_abbrev(UC, LC, 0'2) :-
        uc_code(UC, LC).

%-------------------- Converting a table or a board to a problem
% 
% The predicates below allow converting a table in a quasi-graphical
% format into a problem description. For example, the problem `dense'
% was generated by:
% 
%    | ?- table_to_problem(
%                 [" * : O : # : * * * * = 3 3",
%                  " : : : : : : : : : : : : :",
%                  " * = 3 : # : # # # # : * *",
%                  " * : 3 : # : : : : : : : :",
%                  " * : 3 : # : 3 3 3 : 3 : *",
%                  " : : : : : : : : : : : : :",
%                  " 3 : * : * : # M # # : # #",
%                  " 3 : * : * : : : : : : : :",
%                  " 3 : * : : : # # : # : * *",
%                  " : : : : 3 : : : : # : : :",
%                  " # : # :m3 : 3 3 : # : 3 3",
%                  " : : # : 3 : : : : : : : :",
%                  " o : # : 3 : * * * :o3 : 3"],
%                           P, []), write(P).

% :- export table_to_problem/3.
% The problem corresponding to table Tbl is TDescr, considering options Opts.
% The only option of interest is 'v1', to generate a version 1 TDescr.
% !!!! Warning: no checks are made whether Tbl is valid !!!!
% :- pred table_to_problem(table::in, torp_descr::out).
table_to_problem(Tbl, TDescr, Opts) :-
        table_to_board(Tbl, Bd),
        board_to_problem(Bd, TDescr, Opts).

% :- export board_to_problem/3.
% board_to_problem(Bd, TDescr): The torpedo problem corresponding to
% board Bd is TDescr, considering options Opts.
% The only option of interest is 'v1', to generate a version 1 TDescr.
% !!!! Warning: no checks are made whether Bd is valid !!!!
% :- pred board_to_problem(board1::in, torp_descr::out).
board_to_problem(Bd, TDescr, Opts) :-
        findall(I-J, matrix_elem(I, J, Bd, =), KnownEs),
        board_to_colours(Bd, 1, ClrDescrs),
	TDescr0 = descr(KnownEs,ClrDescrs),
	adjust_torp_descr_for_v1(TDescr0, TDescr, Opts).

% board_to_colours(Bd, Clr, ClrDescrs): The list of colour_descrs
%       corresponding to colours Clr, Clr+1, ... on board Bd is ClrDescrs.
% :- pred board_to_colours(board1::in, colour::in, colours_descr::out).
board_to_colours(Bd, Clr, [ClrDescr|ClrDescrs]) :-
        board_to_colour(Bd, Clr, ClrDescr), !,
        Clr1 is Clr+1,
        board_to_colours(Bd, Clr1, ClrDescrs).
board_to_colours(_, _, []).

% type ship_counts = list(int).

% The type ship_counts is a list of form: [All|Longs], used for counting
% various formations of a given colour. In this operation, which we will
% call (ship count) incrementing, All counts all pieces of a given colour,
% while Longs is a list of counts of ships of length 2, 3, etc.

% board_to_colour(Bd, Clr, ClrDescr): There are pieces of colour Clr on
% board Bd, and the colour_descr corresponding to colour Clr is ClrDescr.
% :- pred board_to_colour(board1::in, colour::in, colour_descr::out).
board_to_colour(Bd, Clr, c(ShCnts,RowOccs,ClmOccs,KnownPs)) :-
        findall(t(I,J,P), (matrix_elem(I, J, Bd, P-Clr), P\==x), KnownPs),
        board_rows_count(Bd, Clr, [0], ShCnts0, RowOccs), % count rows
        transpose(Bd, TBd),
        board_rows_count(TBd, Clr, ShCnts0, ShCnts1, ClmOccs), % count columns
	adjust_ship_counts(ShCnts1, ShCnts).

adjust_ship_counts(ShCnts, [Ones|ShCntsLongs]) :-
	ShCnts = [All|ShCntsLongs],
        All > 0, % All = twice the number of all pieces of colour Clr
	length(ShCnts, ShLenMax),
	sequence(2, ShLenMax, Nums),
	scalar_product(Nums, ShCntsLongs, SProd),
        Ones is All//2-SProd.

% board_rows_count(Rows, Clr, ShCnts0, ShCnts, Occs): ShCnts0, 
% incremented with the counts of colour Clr within rows Rows, gives
% ShCnts. Occs is a list of occurrences of colour Clr in each row of Rows.
% :- pred board_rows_count(list(row1)::in, colour::in, ship_counts::in,
%                          ship_counts::out, occs::out).
board_rows_count([], _, ShCnts, ShCnts, []).
board_rows_count([Row|Rows], Clr, ShCnts0, ShCnts, [Occ|Occs]) :-
        board_row_count(Row, Clr, ShCnts0, ShCnts1),
        ShCnts0 = [ShCnt0|_],
        ShCnts1 = [ShCnt1|_],
        Occ is ShCnt1-ShCnt0,
        board_rows_count(Rows, Clr, ShCnts1, ShCnts, Occs).

% board_row_count(Row, Clr, ShCnts0, ShCnts): ShCnts0, incremented with the
% counts of colour Clr within row Row, gives ShCnts.
% :- pred board_row_count(row1::in, colour::in, ship_counts::in,
%                                               ship_counts::out).
board_row_count([], _, ShCnts, ShCnts).
board_row_count([F|Fs], Clr, ShCnts0, ShCnts) :-
        formation_count(F, Fs, Clr, ShCnts0, ShCnts1, Fs1),
        board_row_count(Fs1, Clr, ShCnts1, ShCnts).

% formation_count(F, Fs0, Clr, ShCnts0, ShCnts, Fs): If there is a
% formation of colour Clr at the beginning of row [F|Fs0], then
% ShCnts0, incremented with the counts of colour Clr for this formation,
% gives ShCnts, and Fs is the rest of the row. Otherwise ShCnts =
% ShCnts0 and Fs = Fs0.
% :- pred formation_count(field1::in, row1::in, colour::in, ship_counts::in,
%			  ship_counts::out, row1::out).
formation_count(_-Clr, Fs0, Clr, ShCnts0, ShCnts, Fs) :- !,
        formation_length(Fs0, Clr, 1, I, Fs),
        formation_length_count(I, ShCnts0, ShCnts).
formation_count(_, Fs, _, ShCnts, ShCnts, Fs).

% formation_length(Fs0, Clr, I0, I, Fs): There is a formation of length I-I0
% at the beginning of row Fs0 and the rest of the row is Fs.
% :- pred formation_length(row1::in, colour::in, int::in, int::out, row1::out).
formation_length([_-Clr|Fs0], Clr, I0, I, Fs) :- !,
        I1 is I0+1,
        formation_length(Fs0, Clr, I1, I, Fs).
formation_length(Fs, _, I, I, Fs).

% formation_length_count(I, ShCnts0, ShCnts): ShCnts0, incremented with the
% counts of a formation of length I, gives ShCnts.
% :- pred formation_length_count(int::in, ship_counts::in, ship_counts::out).
formation_length_count(I, [ShCnt0|ShCnts0], [ShCnt|ShCnts]) :-
        ShCnt is ShCnt0+I, I1 is I-1,
	(   I1 = 0 -> ShCnts = ShCnts0
        ;   replace_nth(I1, ShCnts0, ShCntI0, 0, ShCnts, ShCntI),
	    ShCntI is ShCntI0+1
	).

%-------------------- Version 1 compatibility

% TDescr is the variant of TDescr0 appropriate for version indicated by Opts.
% (Presence of 'v1' indicates version 1.)
% :- pred adjust_ship_counts(torp_descr::in, torp_descr::out, options::in).
adjust_torp_descr_for_v1(TDescr0, TDescr, Opts) :-
	member(v1, Opts), !,
	TDescr0 = descr(KE,CDs0),
	TDescr = descr(KE,CDs),
	adjust_colours_descr_for_v1(CDs0, CDs).
adjust_torp_descr_for_v1(TDescr, TDescr, _Opts).

% adjust_colours_descr_for_v1(CDs0, CDs): CDs is the variant of CDs0,
% compatible with version 1.
% :- pred adjust_colours_descr_for_v1(colours_descr::in, colours_descr::out).
adjust_colours_descr_for_v1([], []).
adjust_colours_descr_for_v1([c(SCs0,ROs,COs,KPs)|CDs0],
			    [c(Pieces,ROs,COs,KPs)|CDs]) :-
	replace_nth(4, SCs0, ShCnt4, 0, SCs, ShCnt4),
	% to make the ship counts list at least 4 long
	Pieces =.. [pieces|SCs],
	adjust_colours_descr_for_v1(CDs0, CDs).


%-------------------- Collapsing a colours descr

% collapsed_colours_descr(CDs, CD) :- CD is the combination of cues in CDs.
% :- pred collapsed_colours_descr(colours_descr::in, colour_descr::out).
collapsed_colours_descr([CD0|CDs], CD) :-
	colour_descrs_add(CDs, CD0, CD).

% colour_descrs_add(CDs, CD0, CD): CD is the combination of cues in [CD0|CDs]. 
% :- pred colour_descrs_add(colours_descr::in, colour_descr::in,
%                           colour_descr::out).
colour_descrs_add([], CD, CD).
colour_descrs_add([CD1|CDs], CD0, CD) :-
        colour_descr_add(CD0, CD1, CD2),
        colour_descrs_add(CDs, CD2, CD).

% colour_descr_add(CD1, CD2, CD): CD is the combination of cues in CD1 and CD2.
% :- pred colour_descr_add(colour_descr::in, colour_descr::in,
%                          colour_descr::out).
colour_descr_add(c(SC1,RO1,CO1,KP1), c(SC2,RO2,CO2,KP2), c(SC,RO,CO,KP)) :-
        list_add(SC1, SC2, SC),
        list_add(RO1, RO2, RO),
        list_add(CO1, CO2, CO),
        append(KP1, KP2, KP).

