% File utils.pl 
% General purpose utilities
% Version 2.0, July 16, 1999
% NLP course, 1999
% (c) Peter Szeredi

:- module(utilities, [
		      % Matrix operations:
		      transpose/2,
		      matrix_prepend/3,
		      matrix_elem/4,
		      matrix_of/4,
		      matrix_of_vars/3,
		      % List operations:
		      append_lists/2,
		      list_of/3,
		      list_add/3,
		      replace_nth/5,
		      replace_nth/6,
		      sequence/3,
		      scalar_product/3,
		      % Character codes:
		      digit_code/1,
		      lc_code/1,
		      uc_code/2,
		      % Miscellaneous:
		      between/3,
		      with_output_to_compiler/1]).
	      
:- use_module(library(lists)).
:- use_module(library(system)).


%-------------------- Matrix operations

% :- type matrix(Type) == list(list(Type)).
% A matrix is represented by a list of rows, which are lists of
% elements. All rows must be of equal length.

% transpose(M, TM): TM is the transposition of nonempty matrix M. 
% :- pred transpose(matrix(Type)::in, matrix(Type)::out).
transpose(Matrix0, [Clm|Clms]) :-
        matrix_prepend(Clm, Matrix1, Matrix0), !,
        transpose(Matrix1, Clms).
transpose(_, []).

% matrix_prepend(Clm, Matrix, ExtMatrix): The column Clm, added to Matrix
% as its first column results in extended matrix ExtMatrix. Can also be
% used for splitting ExtMatrix into a first column Clm and a remaining
% matrix Matrix.
% :- pred matrix_prepend(list(Type), matrix(Type), matrix(Type)).
% :- mode matrix_prepend(in, in, out).
% :- mode matrix_prepend(out, out, in).
matrix_prepend([], [], []).
matrix_prepend([Elem|Clm], [Row|Rows], [ExtRow|ExtRows]) :-
	ExtRow = [Elem|Row],
	matrix_prepend(Clm, Rows, ExtRows).

% matrix_elem(I, J, M, E): E is the element of matrix M in row I, column J.
% :- pred matrix_elem(int::in, int::in, matrix(T)::in, T::out).
matrix_elem(RowI, ClmI, Matrix, Elem) :-
        nth(RowI, Matrix, Row),
        nth(ClmI, Row, Elem).

% Matrix is a matrix of size RowNum x ClmNum, with all elements uninstantiated.
% :- pred matrix_of_vars(int::in, int::in, matrix(T)::out).
matrix_of_vars(0, _ClmNum, []) :- !.
matrix_of_vars(RowNum, ClmNum, [Row|Rows]) :-
	length(Row, ClmNum),
	RowNum1 is RowNum-1,
	matrix_of_vars(RowNum1, ClmNum, Rows).

% Matrix is a matrix of size RowNum x ClmNum, with all elements equal to E.
% :- pred matrix_of(int::in, int::in, T::in, matrix(T)::out).
matrix_of(RowNum, ClmNum, E, Matrix) :-
	list_of(ClmNum, E, Row),
	list_of(RowNum, Row, Matrix).


%-------------------- List operations.

% append_lists(LL, FL): FL is the result of appending all elements of LL.
% :- pred append_lists(list(list(T))::in, list(T)::out).
append_lists([], []).
append_lists([L|LL], FL) :-
	append(L, FL1, FL),
	append_lists(LL, FL1).

% list_of(N, E, L): L is a list of length N with all elements being E.
% :- pred list_of(int::in, T::in, list(T):: out).
list_of(0, _, []) :- !.
list_of(N, E, [E|L]) :-
	N > 0, N1 is N-1,
	list_of(N1, E, L).

% list_add(L0, L1, L): L's length is the maximum of lengths of L0 and L1.
% The Ith elem of L is the sum of Ith elems of L0 and L1. A nonexistent elem
% is assumed to have value 0.
% :- pred list_add(list(int)::in, list(int)::in, list(int)::out).
list_add([X|Xs], [Y|Ys], Zs) :- !,
	Z is X+Y,
	Zs = [Z|Zs0],
	list_add(Xs, Ys, Zs0).
list_add(Xs, [], Zs) :- !, Zs = Xs.
list_add([], Ys, Ys).

% replace_nth(N, L0, X0, L1, X1): Nth element of L0 is X0, Nth element
% of L1 is X1, and all other elements of L0 and L1 are the same
% :- pred replace_nth(int::in, list(T)::in, T::out, list(T)::out, T::out).
replace_nth(1, [X0|L0], X0, [X1|L0], X1) :- !.
replace_nth(N, [X|L0], X0, [X|L], X1) :-
        N1 is N-1,
        replace_nth(N1, L0, X0, L, X1).

% replace_nth(N, L0, X0, E, L1, X1): If N =< length(L0), then the same as
% replace_nth(N, L0, X0, L1, X1). Otherwise L0 is appended N-length(L0)
% elements with value E, and replace_nth/5 is applied to the resulting list.
% :- pred replace_nth(int::in, list(T)::in, T::out, T::in,
%                              list(T)::out, T::out).
replace_nth(1, [X0|L0], X0, _E, [X1|L0], X1) :- !.
replace_nth(N, [], X0, E, [X|L], X1) :-
	replace_nth(N, [E], X0, E, [X|L], X1).
replace_nth(N, [X|L0], X0, E, [X|L], X1) :-
        N1 is N-1,
        replace_nth(N1, L0, X0, E, L, X1).

% sequence(From, To, Seq): Seq is the sequence of consecutive integers
% between From and To.
% :- pred sequence(int::in, int::in, list(int)::out).
sequence(From, To, [From|Rest]) :-
	From =< To, !, From1 is From+1,
	sequence(From1, To, Rest).
sequence(_, _, []).

% SProd is the scalar product of lists Xs and Ys.
% :- pred scalar_product(list(int)::in, list(int)::in, int::out).
scalar_product(Xs, Ys, SProd) :-
	scalar_product(Xs, Ys, 0, SProd).

% scalar_product(Xs, Ys, SProd0, SProd): SProd-SProd0 is the scalar product
% of lists Xs and Ys.
% :- pred scalar_product(list(int)::in, list(int)::in, int::in, int::out).
scalar_product([], [], SProd, SProd).
scalar_product([X|Xs], [Y|Ys], SProd0, SProd) :-
	SProd1 is SProd0+X*Y,
	scalar_product(Xs, Ys, SProd1, SProd).

%-------------------- Character codes

% D is a digit code.
% :- pred digit_code(int::in).
digit_code(D) :-
        0'0 =< D, D =< 0'9.


% LC is a code of a lower case letter.
% :- pred lc_code(int::in).
lc_code(LC) :-
        0'a =< LC, LC =< 0'z.

% UC is a code of an upper case letter, and LC is the corresponding lower
% case letter.
% :- pred lc_code(int::in, int::out).
uc_code(UC, LC) :-
        0'A =< UC, UC =< 0'Z,
        LC is UC-0'A+0'a.


%-------------------- Miscellaneous

% between(I, J, X): I =< X, X =< J. Capable of enumerating all such Xs.
% :- pred between(int::in, int::in, int::out).
between(I, J, I) :- I =< J.
between(I, J, X) :-
	I < J, I1 is I+1,
	between(I1, J, X).

:- meta_predicate	
	with_output_to_compiler(:).

% Run Goal with the 'current_output' output stream set to a file and
% and after successful completion of Goal feed this file to the compiler.
% :- pred with_output_to_compiler(any::in).
with_output_to_compiler(Goal) :-
	tmpnam(FileName),
	open(FileName, write, S), set_output(S),
	call(Goal),
	close(S),
	Goal = M:_,
	prolog_flag(redefine_warnings, Old, off),
	asserta(user:portray_message(informational, _)),
	compile(M:FileName),
	retract(user:portray_message(informational, _)),
	prolog_flag(redefine_warnings, _, Old),
	delete_file(FileName).

